/* header for peace_pad routines
*
*  sjs
*  22 Feb 2001
*   5 April 2001 revised#
*     May 2001 Re-organised Qpeace files
*
*/

#ifndef _QPEACE_PAD
#define _QPEACE_PAD

#ifdef __cplusplus
extern "C"{
#endif


#include "Qidfs_utils.h"
#include "Qsweep_spin.h"
#include "PGSpectro.h"
#include "Qpeace_err_diag.h"

#define MAX_P_ANGLES 13
#define PAD_HAR_HALF_SPIN_MILLIS  1500   
             /* skip at least 1.5 secs to find 2nd half of PAD in HAR */

typedef enum PARALLEL_FIELD_HALF_SPIN_E{
   PARALLEL_FIRST_HALF = 0,
   PARALLEL_SECOND_HALF
} PARALLEL_FIELD_HALF_SPIN;

typedef enum CLOCK_DIRECTION_E{
  COUNTERCLOCKWISE = -1,
  CLOCKWISE        =  1,
  CLOCKSTEPERROR   =  0
} CLOCK_DIRECTION;

/* prototypes */

/* High level routines */

SDDAS_SHORT GetPeacePAD(IDFS_LINEAGE *lineage, IDFS_TIME *btime,
   BIN_UNITS bin_units, PSD_UNITS psd_units, 
   T_OR_F do_pa, PEACE_SWEEP *swpp);
   
SDDAS_SHORT GetNPeacePADs(
        IDFS_LINEAGE *lngpadp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *npads,
	PEACE_SWEEP ***swppp);
   
/* read/fill routine */
		      
SDDAS_SHORT init_pad(PEACE_SWEEP *swpp, IDFS_LINEAGE *lngp, 
    struct idf_data *idfp, SDDAS_ULONG dkey,
    float b_az, PARALLEL_FIELD_HALF_SPIN *b_para_half_spin, int *isen_init);
    
SDDAS_SHORT  fill_pad(PEACE_SWEEP *swpp,
                      IDFS_LINEAGE *lineage,
		      struct idf_data *idf_padp,
		      SDDAS_ULONG dkey_pad, 
		      BIN_UNITS bin_units,
                      PSD_UNITS psd_units,
		      T_OR_F do_pa,
		      PARALLEL_FIELD_HALF_SPIN b_half_spin,
		      int isen);

int expected_first_polbin(int b_polbin_iel, IDFS_LINEAGE *lngp, 
                             PEACE_SWEEP *swpp);

/* B-field routines */    

SDDAS_SHORT get_b_az(SDDAS_ULONG dkey, struct idf_data *idfp, 
                      IDFS_LINEAGE *lineage, float *b_az, int *b_pol);
       
SDDAS_SHORT set_b_para_half_spin(float b_az, 
         IDFS_LINEAGE *lngp, PARALLEL_FIELD_HALF_SPIN *b_para_half_spin);

/* Sorting routines */

int compute_clockbin(int sensor, int num_sensors, int swp_half, int n_pa);

SDDAS_SHORT sort_pa_bins(int *inbins, int *outbins, int pa_dim,
    PARALLEL_FIELD_HALF_SPIN bpara_half, int b_polbin);

int exchange_b_tip_tail(int *bpol, PARALLEL_FIELD_HALF_SPIN *bpara_half, 
                    int n_sensors);

CLOCK_DIRECTION clockstep(int *clockbins, int *startbin, int nbins, 
                   int tries_to_go);

PEACE_SWEEP* PADLiouvilleMap (PEACE_SWEEP *swp1, float B1, float B2, 
  float epot, int n_double_theta);		   


PEACE_SWEEP *PADvpara_boost(PEACE_SWEEP *swpin, float vparaboost,
   int thetafactor, int levelfactor);

PEACE_SWEEP *sweep_double_sensors(PEACE_SWEEP *swpin);
PEACE_SWEEP *sweep_double_levels(PEACE_SWEEP *swpin);		   

float sinsq_map (float e1, float e2, float b2ob1, float thin);

float cyclic360abs_diff(float x1, float x2);

float cyclic360pos_diff(float xlo, float xhi);


#ifdef __cplusplus
}
#endif


#endif  /* end ifndef _PEACE_PAD */
