 
/* Qpeace_plot routines 
*
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  May 2001 - re-organised Qpeace routines and created this module from others
*  Sep 2001  added include float.h to pick up FLT_MIN, etc. for Mandrake 8
*
*  Routines for plotting PEACE sweeps and spins
*  Uses generic plotting routines found in PGSpectro
*
*/

#ifndef _QPEACE_PLOT
#define _QPEACE_PLOT

#ifdef __cplusplus
extern "C"{
#endif

#include "Qidfs_utils.h"
#include "Qsweep_spin.h"
#include "PGSpectro.h"
#include "Qpeace_err_diag.h"
#include <float.h>


#define NCIRCLES 4     /* number of dashed guide circles in sweep plot */


typedef struct SWEEP_WHEEL_OPTIONS_S{
   PGSpectro_col color_table;
   PGSpectro_linlog zlinlog;
   PGSpectro_linlog rlinlog;
   Yes_or_No zautoscale, rautoscale,skip_missing;
   float zmin, zmax, rmin, rmax;
   float title_just;
   char* pg_device;
} SWEEP_WHEEL_OPTIONS;

typedef struct SWEEP_CUT_OPTIONS_S{
   int nswps; 
   QPGLine_colour *colours;      /* one for each swp */
   int ncuts;                    /* for Sweep_cuts_plot each sweep is cut */
                                 /* ncuts times; for Sweep_n_cuts, each sweep*/
				 /* is cut once and all arrays are [nswps] */
   QPGLinestyle   *linestyles;   /* one for each cut */
   float linethickness;          /* only 1; multiple of default */
   PGSpectro_linlog xlinlog;
   PGSpectro_linlog ylinlog;
   Yes_or_No yautoscale, xautoscale;
   float xmin, xmax, ymin, ymax;
   char* pg_device;
} SWEEP_CUT_OPTIONS;


PGSpectro_ret Sweep_wheel_plot(PEACE_SWEEP *swpp,
               SWEEP_WHEEL_OPTIONS *swop);

void Set_sweep_wheel_options_default(SWEEP_WHEEL_OPTIONS *swop);


PGSpectro_ret PlotTwoSweeps(PEACE_SWEEP *swp_pri, PEACE_SWEEP *swp_sec,
       SWEEP_WHEEL_OPTIONS *swop);


PGSpectro_ret Qjas_plot( PEACE_SWEEP *swps[], /* pointer to nswp sweep ptrs */
                   int nswp,           /* number of sweeps */
		   PGSpectro *pgspec,  /* brings in various settings */
		   IDFS_TIME idf_start,/* plot time range */
		   IDFS_TIME idf_end);

void qjas_time_init(IDFS_TIME *idf_startp, IDFS_TIME *idf_endp,
                    char *date_start_str, char *date_end_str,
		    PGSpectro *pgs, double *epoch_midnight);

void Qjas_az_panel(PGSpectro *pgs, PEACE_SWEEP **swps);

void Qjas_set_vpt_win_from_pgs(PGSpectro *pgs);

PGSpectro_ret Qjas_lev_panel(PGSpectro *pgs, PEACE_SWEEP **swps, int lev);

PGSpectro_ret Qjas_fill_angle_t_array(
               float *z,   /* array z[nswp][nang] */
	       int nswp,   /* number. of swps = no. of timetags */
	       int nang,   /* number of angle pts */
	       PEACE_SWEEP **sweeps,  /* pointer to array of nswp sweep ptrs */
               int lev);  /* energy level to be filled */
	       
PGSpectro_ret malloc_spectro_arrays(PGSpectro *pgs, int nx, int ny);

void free_spectro_arrays(PGSpectro *pgs);

PGSpectro_ret Sweep_cuts_plot (PEACE_SWEEP **swpp, int nswps, 
                   float *thetas, int ncuts, 
		    SWEEP_CUT_OPTIONS *scutopt);

PGSpectro_ret Set_Sweep_cut_options_default(SWEEP_CUT_OPTIONS *scopp);

PGSpectro_ret Sweep_n_cuts_plot (PEACE_SWEEP **swpp, int nswps, 
                   float *thetas, SWEEP_CUT_OPTIONS *scutopt);
		    
#ifdef __cplusplus
}
#endif


#endif  /* end ifndef _QPEACE_PLOT */


