/* plotaz.h ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*     May 2001 up-date after re-organization of Qpeace routines
*     Jan 2003 includes mapping, shifting, ...
*/

#ifndef PLOTAZ_H
#define PLOTAZ_H

#ifndef __cplusplus
#define __cplusplus
#endif


#include <qvbox.h>
#include <qhbox.h>
#include <qcombobox.h>
#include <qgroupbox.h>
#include <qtextbrowser.h>
#include <qpushbutton.h>
#include <qstring.h>
#include <qlabel.h>
#include <qframe.h>
#include <qlineedit.h>
#include <qapplication.h>
#include <qdialog.h>
#include <qtooltip.h>

#include "Qsweep_spin.h"
#include "Qpeace_pad.h"
#include "Qidfs_utils.h"
#include "Qpeace_plot.h"

#define PLOTAZ_VERSION_STR "V2.67"
#define PLOTAZ_DEFAULT_XWIN_STR "53"
#define N_MAX_CUTS 6
#define PLOTAZ_LIOUVILLE_UNSHIFT_DOUBLE 7
#define PLOTAZ_LIOUVILLE_SHIFTED_DOUBLE 5
#define PLOTAZ_SHIFT_THETA_FAC 2
#define PLOTAZ_SHIFT_LEVEL_FAC 2

class QLabel;

static const char *project_strs[] = {"CLUSTERII", "DOUBLESTAR", "UNKNOWN", 0};
//static const char *mission_strs[] = {"C1", "C2", "C3", "C4", "TC-1", "TC-2",
//                   "FLIGHTSPARE", 0};  // done in Qidfs_utils.c 
static const char *exp_strs[] = {"PEACE", 0};
static const char *inst_strs[] = {"PAD", "SPINPAD", "3DR", "3DF", "LER", 
         "CORE", "NOI", "MODE", "3DX1", "3DX2", "HSKP", 0};
static const char *virins_strs[] = {
      "CPPADH" , "CPPADHF" , "CPPADL" , "CPPADLF",           // PAD
      "LSPAD", "HSPAD",                                      // SPINPAD
      "CP3DRH" , "CP3DRHF" , "CP3DRL" , "CP3DRLF",           // 3DR
      "CP3DFH" , "CP3DFHF" , "CP3DFL" , "CP3DFLF",           // 3DF
      "CPLERH" , "CPLERHF" , "CPLERL" , "CPLERLF",           // LER
      "CPX1L" , "CPX1H" , "CPXP1L" , "CPXP1H" , "CPXE1L" , "CPXE1H" , "CPX1LF",
         "CPX1HF" , "CPXP1LF" , "CPXP1HF" , "CPXE1LF" , "CPXE1HF",  // 3DX1
      "CPX2L" , "CPX2H" , "CPXP2L" , "CPXP2H" , "CPXE2L" , "CPXE2H" , "CPX2LF",
         "CPX2HF" , "CPXP2LF" , "CPXP2HF" , "CPXE2LF" , "CPXE2HF",  // 3DX2
       "DP3DF",0};


#define N_POSSIBLE_CUT_SOURCES 11   /* includes NO_SOURCE */

typedef enum SWEEP_SOURCE_E{
    NO_SOURCE = 0,
    PRIMARY,
    SHIFTED_PRIMARY,
    MAPPED_PRIMARY,
    SHIFTED_THEN_MAPPED_PRIMARY,
    SECONDARY,
    SHIFTED_SECONDARY,
    MAPPED_OVER_SECONDARY,
    SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY,
    MAPPED_MINUS_SECONDARY,
    SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY
} SWEEP_SOURCE;
    
typedef enum PA_CHOICE_E{
    NO_BINNING = 0,
    PA_ANGLES,
    PA_COSINES
} PA_CHOICE;


class LineageBox : public QVBox
{
    Q_OBJECT

public:
    LineageBox (QWidget *parent = 0, const char *name = 0);
    QComboBox *proj, *miss, *exp, *inst, *virins;

protected:
};

class SweepBox : public QGroupBox
{
    Q_OBJECT

public:
    SweepBox (int strips = 1, Orientation=Horizontal,
                    QWidget *parent = 0, const char *name = 0);
    QLabel *project, *mission, *exp, *inst, *virins;
    QComboBox *pachoice;
    QLineEdit *npa;
    QLineEdit *year, *doy, *hr, *min, *sec, *azslice;
    void SetLineage(LineageBox *);

protected:
    QLabel *lineage_lbl, *time_lbl, *slice_lab;
    // more to go here to set time and az slice

protected slots:
};


class Sources : public QComboBox
{
   Q_OBJECT
   
public:
     Sources (QWidget *parent = 0, const char *name =0);

protected:
protected slots:

};

class Pangles : public QLineEdit
{
   Q_OBJECT
   
public:
   Pangles (QWidget *parent = 0, const char *name =0);

protected:
protected slots:

};

class Linestys : public QComboBox
{
   Q_OBJECT
   
public:
     Linestys (QWidget *parent = 0, const char *name =0);

protected:
protected slots:

};


class Colours : public QComboBox
{
   Q_OBJECT
   
public:
     Colours (QWidget *parent = 0, const char *name =0);

protected:
protected slots:

};

class Cuts_Selector : public QVBox
{
   Q_OBJECT
   
public:
   QLabel  *srclab;
   Sources *source;
   Pangles *pangle;
   Linestys *linesty;
   Colours  *colour;      
   Cuts_Selector (int icut, QWidget *parent = 0, const char *name = 0);
};

class Helpbwin : public QVBox
{
    Q_OBJECT

public:
      Helpbwin (QWidget *parent = 0, const char *name = 0);
      ~Helpbwin();

protected:

protected slots:

};

class Slicewin : public QDialog
{
    Q_OBJECT

public:

    QLineEdit *b1, *b2, *pot, *v1, *v2;
    Sources *lwheel, *rwheel;
    Cuts_Selector *cuts[N_MAX_CUTS];
    QComboBox  *energy_units, *energy_linlog, *energy_scale;
    QComboBox  *psd_units, *psd_linlog, *psd_scale, *plot_col, *plot_dev;
    QLineEdit  *emin, *emax, *psdmin, *psdmax, *plot_win_edit, *plot_file_edit;

    Slicewin (QWidget *parent =0, const char *name = 0);
    ~Slicewin();

protected:

protected slots:
      void slotDismiss(void);
      void slotRun(void);

private:
      float bprim, bsec, potential,vprim, vsec;
      float thetas[N_MAX_CUTS];
      SWEEP_SOURCE lwhsrc, rwhsrc, cutsrc[N_MAX_CUTS];
      QPGLine_colour cutcol[N_MAX_CUTS];
      QPGLinestyle cutlsty[N_MAX_CUTS];
      PEACE_SWEEP *swpp[N_MAX_CUTS];
      PEACE_SWEEP *swposs[N_POSSIBLE_CUT_SOURCES];
      int nswps, nwheels, swpcount[N_POSSIBLE_CUT_SOURCES];
      SWEEP_CUT_OPTIONS sweep_cut_options, *scop;
      SWEEP_WHEEL_OPTIONS sweep_wheel_options, *swop; 
      PGSpectro_ret pg_ret_val;

};

class PlotAzBox : public QVBox
{
    Q_OBJECT

public:
    PlotAzBox( QWidget *parent = 0, const char *name = 0 );
    LineageBox *lineage;
    SweepBox *prim_sweep, *sec_sweep;
    QComboBox  *energy_units, *energy_linlog, *energy_scale;
    QComboBox  *psd_units, *psd_linlog, *psd_scale, *plot_col, *plot_dev;
    QLineEdit  *emin, *emax, *psdmin, *psdmax, *plot_win_edit, *plot_file_edit;
    QComboBox  *anal_proc, *anal_linlog, *anal_scale;
    QLineEdit  *analmin, *analmax;
    QPushButton *quit;
    PEACE_SWEEP *swp1, *swp2,*swpar;
    PEACE_PITCH_ANGLE_OPTS paopts1, paopts2;
    IDFS_LINEAGE lng1, lng2;
    IDFS_TIME idft1, idft2;
    BIN_UNITS bin_units;
    PSD_UNITS psdu;
    PGSpectro_ret pg_ret_val;
    SWEEP_WHEEL_OPTIONS swoptions;
    float az1, az2;
    Helpbwin *helpbwin;
    Slicewin *slicewin;


protected:
//    LineageBox *lineage;
//    SweepBox *prim_sweep, *sec_sweep;


protected slots:
    void slotSetPrimarySweepLineage(void);
    void slotSetSecondarySweepLineage(void);
    void slotCopyTime(void);
    void slotRun(void);
    void slotSlice(void);
    void slotHelp(void);

private:

    void transfer_lineage(SweepBox *swpbx, IDFS_LINEAGE *lng);
    void transfer_sweep_time(SweepBox *swpbx, IDFS_TIME *idft);
};


#endif
