/* qjas.h ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*   June 2001  hacked from plotaz.h 
*
*/

#ifndef QJAS_H
#define QJAS_H

#ifndef __cplusplus
#define __cplusplus
#endif

#include <qvbox.h>
#include <qhbox.h>
#include <qcombobox.h>
#include <qgroupbox.h>
#include <qtextbrowser.h>
#include <qpushbutton.h>
#include <qstring.h>
#include <qlabel.h>
#include <qframe.h>
#include <qlineedit.h>
#include <qapplication.h>
#include <qtooltip.h>
#include <qsizepolicy.h>

#include "Qsweep_spin.h"
#include "Qpeace_pad.h"
#include "Qidfs_utils.h"
#include "Qpeace_plot.h"

#define QJAS_VERSION_STR "V2.67"
#define PLOTAZ_DEFAULT_XWIN_STR "53"

class QLabel;

static const char *project_strs[] = {"CLUSTERII", "DOUBLESTAR", "UNKNOWN", 0};
//static const char *mission_strs[] = {"C1", "C2", "C3", "C4", "TC-1", "TC-2",
//                   "FLIGHTSPARE", 0};  // done in Qidfs_utils.c 
static const char *exp_strs[] = {"PEACE", 0};
static const char *inst_strs[] = {"PAD", "SPINPAD", "3DR", "3DF", "LER", 
         "CORE", "NOI", "MODE", "3DX1", "3DX2", "HSKP", 0};
static const char *virins_strs[] = {
      "CPPADH" , "CPPADHF" , "CPPADL" , "CPPADLF",           // PAD
      "LSPAD", "HSPAD",                                      // SPINPAD
      "CP3DRH" , "CP3DRHF" , "CP3DRL" , "CP3DRLF",           // 3DR
      "CP3DFH" , "CP3DFHF" , "CP3DFL" , "CP3DFLF",           // 3DF
      "CPLERH" , "CPLERHF" , "CPLERL" , "CPLERLF",           // LER
      "CPX1L" , "CPX1H" , "CPXP1L" , "CPXP1H" , "CPXE1L" , "CPXE1H" , "CPX1LF",
         "CPX1HF" , "CPXP1LF" , "CPXP1HF" , "CPXE1LF" , "CPXE1HF",  // 3DX1
      "CPX2L" , "CPX2H" , "CPXP2L" , "CPXP2H" , "CPXE2L" , "CPXE2H" , "CPX2LF",
         "CPX2HF" , "CPXP2LF" , "CPXP2HF" , "CPXE2LF" , "CPXE2HF",  // 3DX2
       "DP3DF",0};


typedef enum PA_CHOICE_E{
    NO_BINNING = 0,
    PA_ANGLES,
    PA_COSINES
} PA_CHOICE;

class LineageBox : public QVBox
{
    Q_OBJECT

public:
    LineageBox (QWidget *parent = 0, const char *name = 0);
    QComboBox *proj, *miss, *exp, *inst, *virins;
    QComboBox *pachoice;
    QLineEdit *npa;
    void grab_lineage(IDFS_LINEAGE *lngp);

protected:

};


class TimeBox : public QGroupBox
{
    Q_OBJECT

public:
    TimeBox (int strips = 1, Orientation=Horizontal,
                    QWidget *parent = 0, const char *name = 0);
    QLineEdit *year, *doy, *hr, *min, *sec;
    void GetTimeFromTimeBox(IDFS_TIME *);

protected:
    QLabel  *time_lbl;

protected slots:

};

class Helpbwin : public QVBox
{
    Q_OBJECT
    
public:
      Helpbwin (QWidget *parent = 0, const char *name = 0);
      ~Helpbwin();

protected:

protected slots:

};


class QjasBox : public QVBox
{
    Q_OBJECT

public:
    QjasBox( QWidget *parent = 0, const char *name = 0 );
    LineageBox *lineage;
    TimeBox *starttime, *endtime;
    QComboBox  *energy_units, *energy_bins, *energy_linlog, *energy_scale;
    QComboBox  *psd_units, *psd_linlog, *psd_scale, *plot_col, *plot_dev;
    QLineEdit  *emin, *emax, *psdmin, *psdmax, *plot_win_edit, *plot_file_edit;
    QLineEdit  *nebins;
    QComboBox  *exp_opt;
    QLineEdit  *exp_file_edit;
    QPushButton *quit;
    PEACE_SWEEP **swppp;   /* array of ptrs to sweep structures; array malloced */
    PEACE_PITCH_ANGLE_OPTS paopts;
    PEACE_SWEEP **cong_swps;
    int nswp;
    IDFS_LINEAGE lng1;
    IDFS_TIME idft1, idft2;
    BIN_UNITS bin_units;
    PSD_UNITS psdu;
    PGSpectro_ret pg_ret_val;
    PGSpectro pgspectro;
    Helpbwin *helpbwin;

protected:

protected slots:
    void slotCopyTime(void);
    void slotExp(void);
    void slotPlot(void);
    void slotHelp(void);

private:

};


#endif
