/* Main programme for Liouville Mapping development
*
*  sjs
*  24 April 2001
*  8 May    2001  try out PAD sweep gathering routines
*  26 Oct 2001  hacked from Qjas_pad_plot
*/

#include "Qsweep_spin.h"
#include "Qpeace_pad.h"
#include "Qidfs_utils.h"
#include "Qpeace_plot.h"
 

int main(){

/* INPUTS************************************************************/
IDFS_LINEAGE lineage = { "CLUSTERII",     /* project */
                         "CLUSTER-2",     /* mission */
                         "PEACE",         /* experiment */
                         "PAD",           /* instrument */
                         "CPPADL",        /* vir instr */
			 "",              /* extension */
                         -100             /* dummy version number */
                       };

int syear  = 2001;    int eyear  = 2001;
int sdoy   =   90;    int edoy   =   90;
int shour  =    18;    int ehour  =   18;
int smin   =   2;    int emin   =   2;
int ssec   =    45;    int esec   =   55;
long snano =    0;    long enano =    0;

float B1 = 1.23456, B2  = 2.234560, epot = 20.0 ;
float thetacut[2] = {0., 90.};
SWEEP_CUT_OPTIONS scop;


BIN_UNITS bin_units = BIN_UNITS_KM_S;    /* BIN_UNITS_KM_S or BIN_UNITS_EV; */
PSD_UNITS psd_units = PSD_UNITS_S3_PER_KM6; 
/* PSD_UNITS_CNTS_ACCUM; */ /* or PSD_UNITS_S3_PER_KM6; */
/**********************************************************************/

IDFS_TIME btime,etime;
SDDAS_SHORT ret_val;

IDFS_LINEAGE *lngp;

PGSpectro_ret pgsret;
PEACE_SWEEP sweep, *swp, *mapped_swp;
SWEEP_WHEEL_OPTIONS sweep_options, *swop; 
PEACE_1D_SWEEP_CUT *swpcut;
PEACE_SWEEP *swpp[2];
QPGLine_colour cols[2];
QPGLinestyle linesty[2];

swp = &sweep;
swop = &sweep_options;
lngp=&lineage; 

for(;;){   /* forever loop to test memory */


to_idfs_time( syear, sdoy, shour, smin, ssec, snano, &btime);
to_idfs_time( eyear, edoy, ehour, emin, esec, enano, &etime);

printf("btime = ");print_idfs_time(&btime);
printf("\nlasttime = ");print_idfs_time(&etime);printf("\n");

ret_val = GetPeacePAD(lngp,&btime,bin_units,psd_units, QPTRUE,swp);
if(ret_val != ALL_OKAY){
   printf("GetPeacePAD failed returning %d\n",ret_val);
   exit(-1);
}

printf("Sweep before mapping:\n------------------------------\n");
print_sweep_structure(swp);

mapped_swp = PADLiouvilleMap(swp, B1, B2, epot,6);  /* double thetas 6 times */

if (mapped_swp == NULL){
   printf("null mapped swp rerturned frorm PADLiouvilleMap; exit\n");
   exit(0);
}

printf("Sweep after mapping:\n------------------------------\n");
print_sweep_structure(mapped_swp);


if (cpgopen("?Device ?") <=0){ 
/*if (cpgopen("/xser") <=0){ */
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);
 
Set_sweep_wheel_options_default(swop);

swop->rautoscale=NO;
swop->rmin = 100;
swop->rmax =20000.;
swop->zautoscale=NO;
swop->zmin = .1; swop->zmax = 1.e8;

pgsret=PlotTwoSweeps(swp, mapped_swp, swop);

 
cpgclos();

if (cpgopen("?Device ?") <=0){ 
/*if (cpgopen("/xser") <=0){ */
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);

swpcut = sweep_1D_cut(thetacut[0], mapped_swp);
printf("Cut of mapped sweep at theta = %g\n======================\n",
         thetacut[0]);
print_1D_cut(swpcut);

free_1D_cut_arrays(swpcut);
free(swpcut);

Set_Sweep_cut_options_default(&scop);

swpp[0] = swp;
swpp[1] = mapped_swp;
scop.nswps = 2;
scop.ncuts = 2;
cols[0] = QPGBLACK; cols[1]=QPGRED;
linesty[0] = QPGSOLID;
linesty[1] = QPGDASHED;
scop.colours = cols;
scop.linestyles = linesty;
scop.linethickness = 2.5;

/* Sweep_cuts_plot (swpp, 2, thetacut, 2, &scop); */ /* makes 4 cuts */
Sweep_n_cuts_plot (swpp, 2, thetacut,&scop);          /* makes 2 cuts */
cpgclos();

free_sweep_arrays(swp);
free_sweep_arrays(mapped_swp);

printf("hit return");
getc(stdin);

}  /* end infinite loop */


exit(0);

}
