/* Main programme for vpara boost development
*
*  sjs
*  24 April 2001
*  8 May    2001  try out PAD sweep gathering routines
*  26 Oct 2001  hacked from Qjas_pad_plot
*  nov 2002 hacked from PADLiouville_main.c
*/

#include "Qsweep_spin.h"
#include "Qpeace_pad.h"
#include "Qidfs_utils.h"
#include "Qpeace_plot.h"
 

int main(){

/* INPUTS************************************************************/
IDFS_LINEAGE lineage = { "CLUSTERII",     /* project */
                         "CLUSTER-2",     /* mission */
                         "PEACE",         /* experiment */
                         "PAD",           /* instrument */
                         "CPPADL",        /* vir instr */
			 "",              /* extension */
                         -100             /* dummy version number */
                       };

int syear  = 2001;    int eyear  = 2001;
int sdoy   =   90;    int edoy   =   90;
int shour  =    18;    int ehour  =   18;
int smin   =   2;    int emin   =   2;
int ssec   =    45;    int esec   =   55;
long snano =    0;    long enano =    0;

float vparaboost = 1000. ;
float thetacut[2] = {0., 0.};
SWEEP_CUT_OPTIONS scop;


BIN_UNITS bin_units = BIN_UNITS_KM_S;    /* BIN_UNITS_KM_S or BIN_UNITS_EV; */
PSD_UNITS psd_units = PSD_UNITS_S3_PER_KM6;
/* PSD_UNITS_DIFF_ENERGY_FLUX; */
/* PSD_UNITS_CNTS_ACCUM; */ /* or PSD_UNITS_S3_PER_KM6; */
/**********************************************************************/

IDFS_TIME btime,etime;
SDDAS_SHORT ret_val;

IDFS_LINEAGE *lngp;

PGSpectro_ret pgsret;
PEACE_SWEEP sweep, *swp, *boost_swp;
SWEEP_WHEEL_OPTIONS sweep_options, *swop; 
PEACE_1D_SWEEP_CUT *swpcut;
PEACE_SWEEP *swpp[2];
QPGLine_colour cols[2];
QPGLinestyle linesty[2];

swp = &sweep;
swop = &sweep_options;
lngp=&lineage; 

for(;;){   /* forever loop to test memory */


to_idfs_time( syear, sdoy, shour, smin, ssec, snano, &btime);
to_idfs_time( eyear, edoy, ehour, emin, esec, enano, &etime);

printf("btime = ");print_idfs_time(&btime);
printf("\nlasttime = ");print_idfs_time(&etime);printf("\n");

ret_val = GetPeacePAD(lngp,&btime,bin_units,psd_units, QPFALSE,swp);
if(ret_val != ALL_OKAY){
   printf("GetPeacePAD failed returning %d\n",ret_val);
   exit(-1);
}

printf("Sweep before boosting:\n------------------------------\n");
print_sweep_structure(swp);

boost_swp = PADvpara_boost(swp, vparaboost,2,4);

if (boost_swp == NULL){
   printf("null boost swp rerturned frorm PADvpara_boost; exit\n");
   exit(0);
}

printf("Sweep after boosting:\n------------------------------\n");
print_sweep_structure(boost_swp);


/*if (cpgopen("?Device ?") <=0){ */
if (cpgopen("4/xwin") <=0){ 
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);
 
Set_sweep_wheel_options_default(swop);

swop->rlinlog=PGSPECTRO_LIN;
swop->rautoscale=YES;
swop->rmin = 100;
swop->rmax =20000.;
swop->zautoscale=YES;
/* swop->zmin = .1; swop->zmax = 1.e8; */

pgsret=PlotTwoSweeps(swp, boost_swp, swop);

 
cpgclos();

/* if (cpgopen("?Device ?") <=0){ */
if (cpgopen("5/xwin") <=0){ 
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);

swpcut = sweep_1D_cut(thetacut[0], boost_swp);
printf("Cut of boost sweep at theta = %g\n======================\n",
         thetacut[0]);
print_1D_cut(swpcut);

free_1D_cut_arrays(swpcut);
free(swpcut);

Set_Sweep_cut_options_default(&scop);

swpp[0] = swp;
swpp[1] = boost_swp;
scop.nswps = 2;
scop.ncuts = 2;
cols[0] = QPGBLACK; cols[1]=QPGRED;
linesty[0] = QPGSOLID;
/*linesty[1] = QPGDASHED; */
linesty[1] = QPGSOLID;
scop.colours = cols;
scop.linestyles = linesty;
scop.linethickness = 2.5;
scop.xlinlog = PGSPECTRO_LIN;

/* Sweep_cuts_plot (swpp, 2, thetacut, 2, &scop); */ /* makes 4 cuts */
Sweep_n_cuts_plot (swpp, 2, thetacut,&scop);          /* makes 2 cuts */
cpgclos();

free_sweep_arrays(swp);
free_sweep_arrays(boost_swp);

printf("hit return");
getc(stdin);

}  /* end infinite loop */


exit(0);

}
