/* Main programme for Qjas_plot development
*
*  sjs
*  24 April 2001
*  8 May    2001  try out PAD sweep gathering routines
*  May 2002  developing pitch angle routines
*
*/

#include  "Qpeace_plot.h"
#include  "Qpeace_pad.h" 

int main(){

/* INPUTS************************************************************/
IDFS_LINEAGE lineage = { "CLUSTERII",     /* project */
                         "CLUSTER-1",     /* mission */
                         "PEACE",         /* experiment */
                         "PAD",           /* instrument */
                         "CPPADL",        /* vir instr */
			 "",              /* extension */
                         -100             /* dummy version number */
                       };

int syear  = 2001;    int eyear  = 2001;
int sdoy   =   90;    int edoy   =   90;
int shour  =    18;    int ehour  =   18;
int smin   =   0;    int emin   =   0;
int ssec   =    00;    int esec   =    5;
long snano =    0;    long enano =    0;

BIN_UNITS bin_units = BIN_UNITS_EV;    /* BIN_UNITS_KM_S or BIN_UNITS_EV; */
PSD_UNITS psd_units = PSD_UNITS_CNTS_ACCUM; /* PSD_UNITS_DIFF_ENERGY_FLUX; */
/* PSD_UNITS_CNTS_ACCUM; */ /* or PSD_UNITS_S3_PER_KM6; */
/**********************************************************************/

IDFS_TIME btime,etime;
SDDAS_SHORT ret_val;

IDFS_LINEAGE *lngp;

PGSpectro_ret pgsret;
PEACE_SWEEP **swppp;   /* array of ptrs to sweep structures; array malloced */
PEACE_SWEEP **cong_swps, **pa_swps;
PEACE_PITCH_ANGLE_OPTS paopt;
int i, nswp;
T_OR_F do_pa;
SWEEP_WHEEL_OPTIONS swop;

PGSpectro pgspectro, *pgsp;

pgsp=&pgspectro;
lngp=&lineage; 

do_pa = QPTRUE;

for(;;){   /* forever loop to test memory */


to_idfs_time( syear, sdoy, shour, smin, ssec, snano, &btime);
to_idfs_time( eyear, edoy, ehour, emin, esec, enano, &etime);

printf("btime = ");print_idfs_time(&btime);
printf("\nlasttime = ");print_idfs_time(&etime);printf("\n");

ret_val = GetNPeaceSweeps(lngp,&btime,&etime,bin_units,psd_units, 
                  do_pa, &nswp, &swppp);
if(ret_val != ALL_OKAY){
   printf("GetNPeaceSweeps failed returning %d\n",ret_val);
   exit(-1);
}

for (i=0; i<nswp; i++){
    printf("Sweep %d including pitch angles is\n", i);
    print_sweep_structure(swppp[i]);
}

set_pa_opts_defaults(&paopt);

pa_swps = (PEACE_SWEEP **)malloc(nswp*sizeof(PEACE_SWEEP*));
for(i=0; i<nswp; i++){
    ret_val = rebin_sweep_to_pa(&(swppp[i]), 1, &paopt, &(pa_swps[i]));
    if (ret_val != ALL_OKAY){
       printf("Rebin_sweep_to_pa returned %d\n",ret_val);
       exit(0);
    }
}

for (i=0; i<nswp; i++){
    printf("PA Sweep %d is\n", i);
    print_sweep_structure(pa_swps[i]);
}


 if (cpgopen("?Device ?") <=0){ 
/* if (cpgopen("/xser") <=0){ */
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);

Set_sweep_wheel_options_default(&swop);
PlotTwoSweeps(swppp[0],pa_swps[0],&swop);
printf("hit return");
getc(stdin);
cpgpage();

 
PGSpectro_set_defaults(pgsp);
sweeps_autoscaler(swppp, nswp, &(pgsp->zscalemin), &(pgsp->zscalemax)); 

/* pgsp->zscalemax = 4.e-4; */

cong_swps=NULL;

pgsp->zlinlog = PGSPECTRO_LOG;
pgsret=Qjas_plot(swppp, nswp, pgsp, btime, etime);
if (pgsret == PGSPECTRO_QJAS_INCONGRUENT){
   cong_swps = (PEACE_SWEEP **)malloc(nswp * sizeof (PEACE_SWEEP *));
   for (i=0; i<nswp; i++){
      cong_swps[i] = rebin_sweep(swppp[0], swppp[i]);
   }
   pgsret=Qjas_plot(cong_swps, nswp, pgsp, btime, etime);
}
 
cpgend();

free_n_sweeps(nswp,swppp);
free(swppp);
swppp=NULL;

free_n_sweeps(nswp,pa_swps);
free(pa_swps);
swppp=NULL;

if (cong_swps != NULL){
  free_n_sweeps(nswp,cong_swps);
  free(cong_swps);
  cong_swps = NULL;
}

printf("hit return");
getc(stdin);

}  /* end infinite loop */


exit(0);

}
