/* Main programme for sweep export development
*
*  sjs
*  July 2001 based on QSAS prog example*
*/

#include  "Qpeace_plot.h"
#include  "Qpeace_pad.h"    /* includes Qsweep_spin.h, which includes qie.h */

void QiCDFContentsDump(QiCDFContents *QiCDF);

int main(){

/* INPUTS************************************************************/
IDFS_LINEAGE lineage = { "CLUSTERII",     /* project */
                         "CLUSTER-1",     /* mission */
                         "PEACE",         /* experiment */
                         "3DR",           /* instrument */
                         "CP3DRL",        /* vir instr */
			 "",              /* extension */
                         -100             /* dummy version number */
                       };

int syear  =  2001;    int eyear  =  2001;
int sdoy   =    24;    int edoy   =    24;
int shour  =    07;    int ehour  =    07;
int smin   =    15;    int emin   =    16;
int ssec   =    30;    int esec   =    00;
long snano =     0;    long enano =     0;

BIN_UNITS bin_units = BIN_UNITS_EV;    /* BIN_UNITS_KM_S or BIN_UNITS_EV; */
PSD_UNITS psd_units =  PSD_UNITS_CNTS_ACCUM; /* PSD_UNITS_DIFF_ENERGY_FLUX*/ 
/* PSD_UNITS_CNTS_ACCUM; */ /* or PSD_UNITS_S3_PER_KM6; */
/**********************************************************************/

IDFS_TIME btime,etime;
SDDAS_SHORT ret_val;

IDFS_LINEAGE *lngp;

PGSpectro_ret pgsret;
PEACE_SWEEP **swppp;   /* array of ptrs to sweep structures; array malloced */
PEACE_SWEEP **cong_swps;
int i, npads;
long qi_ret;

QiCDFContents *QiCDF;
QiOptions *swpsQiOptions;

PGSpectro pgspectro, *pgsp;

pgsp=&pgspectro;
lngp=&lineage; 

QiCDF=NULL;
swpsQiOptions = NULL;


cong_swps = NULL;



to_idfs_time( syear, sdoy, shour, smin, ssec, snano, &btime);
to_idfs_time( eyear, edoy, ehour, emin, esec, enano, &etime);

printf("btime = ");print_idfs_time(&btime);
printf("\nlasttime = ");print_idfs_time(&etime);printf("\n");

for(;;){   /* forever loop to test memory */


ret_val = GetNPeaceSweeps(lngp,&btime,&etime,bin_units,psd_units, QPFALSE,&npads, &swppp);
if(ret_val != ALL_OKAY){
   printf("GetNPeaceSweeps failed returning %d\n",ret_val);
   exit(-1);
}

if (cpgopen("?Device ?") <=0){ 

/* if (cpgopen("/xser") <=0){ */
   printf("Failed to open pgplot device successfully\n");
   exit(0);
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);
 
PGSpectro_set_defaults(pgsp);
sweeps_autoscaler(swppp, npads, &(pgsp->zscalemin), &(pgsp->zscalemax)); 

cong_swps=NULL;

pgsp->zlinlog = PGSPECTRO_LOG;
pgsret=Qjas_plot(swppp, npads, pgsp, btime, etime);
if (pgsret == PGSPECTRO_QJAS_INCONGRUENT){
   cong_swps = (PEACE_SWEEP **)malloc(npads * sizeof (PEACE_SWEEP *));
   for (i=0; i<npads; i++){
      cong_swps[i] = rebin_sweep(swppp[0], swppp[i]);
   }
   pgsret=Qjas_plot(cong_swps, npads, pgsp, btime, etime);
}
 
cpgend();



printf("Now try to export sweeps:\n");
ret_val = nsweeps2QiCDFContents( swppp, npads, &QiCDF);
if (ret_val != ALL_OK){
   printf("nsweeps2QiCDFContents returned %d\n", ret_val);
}

swpsQiOptions = QiMakeOptionsObj();

   swpsQiOptions->header = ATTACHED;
/*   swpsQiOptions->f_type = TABULAR;  also DELIMITED <=> pure csv  */
   swpsQiOptions->f_type = EXCHANGE;
   swpsQiOptions->rec_numbering = NUM_OFF;
   swpsQiOptions->priority = REPLACE;  
   swpsQiOptions->object_type = TS;
   swpsQiOptions->attr_delim = ',';
   swpsQiOptions->data_delim = ',';
   swpsQiOptions->fp_display = stdout;

QiCDF->io_f_name = QiNewStr("nswps_out_file");
QiCDF->io_f_path = QiNewStr("./");
QiCDF->io_f_extn = QiNewStr(".qft");

QiCDFContentsDump(QiCDF);

qi_ret = QiWriteCSDSgenCDF(QiCDF,swpsQiOptions);
if (qi_ret != QMW_OK){
   printf("QiWriteCSDSgenCDF returned %ld\n", qi_ret);
}


qi_ret = QiWriteCSDSgenFlat(QiCDF, swpsQiOptions);
if (qi_ret != QMW_OK){
   printf("QiWriteCSDSgenFlat returned %ld\n", qi_ret);
}


QiFreeOptionsObj (swpsQiOptions);

QiFreeCDFContentsObj( QiCDF); 

/*
printf("Done exporting sweeps\nFirst sweep was\n");
print_sweep_structure (swppp[0]);
printf("\n\nLast sweep structure was\n");
print_sweep_structure (swppp[npads-1]);
*/



free_n_sweeps(npads,swppp);
free(swppp);
swppp=NULL;

if (cong_swps != NULL){
  free_n_sweeps(npads,cong_swps);
  free(cong_swps);
  cong_swps = NULL;
}

printf("  File written. Full loop: swps, Opts + Write routine. hit return");
getc(stdin);

}  /* end infinite loop */


exit(0);

}

void QiCDFContentsDump(QiCDFContents *QiCDF){

int i, dim, att, n_total, ival;
char temp_str[500];
float temp_float;
QiCDFVariable *var;


printf("QiCDF Structure has %ld vars and %ld records\n",QiCDF->n_vars,
    QiCDF->n_recs);

for (i=0; i< QiCDF->n_vars; i++){
  var = QiCDF->vardata[i];
  printf("Variable %d is |%s|\n", i, var->name);
  printf("sizeofentry = %ld, variable number = %ld, datatype = %ld\n",
            var->sizeofentry, var->number, var->data_type);
  printf("rec_vary = %ld, max_rec_num = %ld, num_dims = %ld, num_elems = %ld\n",
          var->rec_vary, var->max_rec_num, var->num_dims, var->num_elems); 
  printf("Dimension info:\n");
  n_total = 1;
  for (dim=0; dim < var->num_dims; dim++){
    n_total *= var->dim_sizes[dim];
    printf("Dim %d:  dim_size = %ld, dim_vary = %ld, dim_depend = %s\n",
          dim, var->dim_sizes[dim], var->dim_varys[dim], var->dim_depends[dim]);
  }
  printf("Attribute info: There are %ld variable attrs\n", var->num_v_attrs);
  for (att = 0; att < var->num_v_attrs; att++){
    printf("Attr %d has number %ld and name |%s|\n",att, 
      var->attribute[att].number, var->attribute[att].name);
    printf("data_type = %ld with %ld elements\n",var->attribute[att].data_type,
       var->attribute[att].num_elems);
    if (var->attribute->data_type == CDF_CHAR){
      printf ("CDF_CHAR attribute found: |%s|\n",(char *)var->attribute[att].data);
    }
  }
  switch (var->data_type){
     case CDF_CHAR:
       for (ival = 0; ival < n_total; ival++){
          strncpy(temp_str, var->data + ival*var->num_elems, var->num_elems);
	  *(temp_str+var->num_elems) = '\0';
          printf ("ival = %d, n_elem chars are |%s|\n",ival,temp_str);
       }
       break;
     case  CDF_FLOAT:
       for (ival = 0; ival < n_total; ival ++){
          printf("ival = %d, float data[ival] = %g\n", ival, 
	    ((float *)var->data)[ival]);
       }
       break;
     case  CDF_DOUBLE:
       for (ival = 0; ival < n_total; ival ++){
          printf("ival = %d, double data[ival] = %g\n", ival, 
	    ((double *)var->data)[ival]);
       }
       break;
     default:
       printf("Variable data type not char, float or double\n");
       break;
  } 
         
  printf("\n------------------------------\n");
}
return;

}
