/* Main programme for sweep export development
*
*  sjs
*  July 2002 based on QJAS export, which is more flexible (pitch angles,
*  re-binning in energy, ...)
*
*  Gather PEACE sweep data for all sensors and output to a CDF or Cluster
*  Exchange Format (cef) file as a time series of Energy/angle arrays.
*  Copes with PAD, SPINPAD, and 3D data.
*/

#include  "Qpeace_pad.h"    /* includes Qsweep_spin.h, which includes qie.h */

void QiCDFContentsDump(QiCDFContents *QiCDF);

int main(){

/* INPUTS************************************************************/
IDFS_LINEAGE lineage = { "DOUBLESTAR",     /* project */
                         "TC-1",     /* mission */
                         "PEACE",         /* experiment */
                         "3DF",           /* instrument */
                         "DP3DF",        /* vir instr */
			 "",              /* extension */
                         -100             /* dummy version number */
                       };
/* start and end time; could load btime & etime structs directly;
*  see Qidfs_utils.h for definition of structure. */

int syear  =  2004;    int eyear  =  2004;   
int sdoy   =    81;    int edoy   =    81;
int shour  =    12;    int ehour  =    12;
int smin   =    00;    int emin   =    00;
int ssec   =    00;    int esec   =    10;
long snano =     0;    long enano =     0;

/* uncomment one of the following */
BIN_UNITS bin_units = BIN_UNITS_EV;    
/* BIN_UNITS bin_units = BIN_UNITS_KM_S; */

/* uncomment one of the following */
PSD_UNITS psd_units =  PSD_UNITS_CNTS_ACCUM;
/* PSD_UNITS psd_units = PSD_UNITS_DIFF_ENERGY_FLUX ; */
/* PSD_UNITS psd_units = PSD_UNITS_S3_PER_KM6;        */

char file_name[] = "sweep_export_file";   /* stem for file name */
char path_name[] = "./";                  /* path; ends with slash */
QiOptions_f_type f_type = EXCHANGE;       /* or UNSET -> cdf; */
 /* TABULAR or DELIMITED also available, but restricted to 1-line records,
 *  so not very nice for sweeps                                         */
/**********************************************************************/

IDFS_TIME btime,etime;
SDDAS_SHORT ret_val;

IDFS_LINEAGE *lngp;

PEACE_SWEEP **swppp;   /* array of ptrs to sweep structures; array malloced */
int nswps;
long qi_ret;

QiCDFContents *QiCDF;
QiOptions *swpsQiOptions;

lngp=&lineage; 

QiCDF=NULL;
swpsQiOptions = NULL;

to_idfs_time( syear, sdoy, shour, smin, ssec, snano, &btime);
to_idfs_time( eyear, edoy, ehour, emin, esec, enano, &etime);

print_idfs_lineage(lngp);                      /* diagnostic only */
printf("btime = ");print_idfs_time(&btime);
printf("\nlasttime = ");print_idfs_time(&etime);printf("\n");

/* get the sweeps; QPFALSE turns off pitch_angle gathering */
ret_val = GetNPeaceSweeps(lngp,&btime,&etime,bin_units,psd_units, 
                       QPFALSE, &nswps, &swppp);
if(ret_val != ALL_OKAY){
   printf("GetNPeaceSweeps failed returning %d\n",ret_val);
   exit(-1);
}

/* Put sweep data into QiCDFContents structure */
ret_val = nsweeps2QiCDFContents( swppp, nswps, &QiCDF);
if (ret_val != ALL_OKAY){
   printf("nsweeps2QiCDFContents returned %d\n", ret_val);
}

/* set up export options; most are ignored if exporting to cdf */
swpsQiOptions = QiMakeOptionsObj();
   swpsQiOptions->header = ATTACHED;   /* or DETACHED                         */
   swpsQiOptions->f_type = f_type;     /* EXCHANGE or TABULAR or DELIMITED    */
   swpsQiOptions->rec_numbering = NUM_OFF;  /* or NUM_ON                      */
   swpsQiOptions->priority = REPLACE;  /* or WARN                             */
   swpsQiOptions->object_type = TS;    /* time series; also DS = Data Series  */
   swpsQiOptions->attr_delim = ',';    /* separator for attributes            */
   swpsQiOptions->data_delim = ',';    /* separator for items in data         */
   swpsQiOptions->rec_end = '$' ;      /* end of record marker                */
   swpsQiOptions->row_end = '\n';      /* end of row; multi-row records       */
   swpsQiOptions->time_sep_attr = 'T'; /* ISO time; T in middle for attributes*/
   swpsQiOptions->time_sep_data = 'T'; /* ditto for data                      */
   swpsQiOptions->fp_display = stdout; /* file pointer for logs, etc          */

QiCDF->io_f_name = QiNewStr(file_name);
QiCDF->io_f_path = QiNewStr(path_name);

/* diagnostic:
QiCDFContentsDump(QiCDF);
*/

switch(f_type){
  case UNSET :
     /* cdf file writing */
     qi_ret = QiWriteCSDSgenCDF(QiCDF,swpsQiOptions);
     if (qi_ret != QMW_OK){
	printf("QiWriteCSDSgenCDF returned %ld\n", qi_ret);
     }
     break;
  case EXCHANGE: default:
     /* cef file writing; we could also write "delimited" or "tabular" */
     qi_ret = QiWriteCSDSgenFlat(QiCDF, swpsQiOptions);
     if (qi_ret != QMW_OK){
	printf("QiWriteCSDSgenFlat returned %ld\n", qi_ret);
     }
     break;
}

/* tidy up */
QiFreeOptionsObj (swpsQiOptions);
QiFreeCDFContentsObj( QiCDF); 
free_n_sweeps(nswps,swppp);
free(swppp);
swppp=NULL;

printf("File written.\n");

exit(0);

}

void QiCDFContentsDump(QiCDFContents *QiCDF){

int i, dim, att, n_total, ival;
char temp_str[500];
QiCDFVariable *var;


printf("QiCDF Structure has %ld vars and %ld records\n",QiCDF->n_vars,
    QiCDF->n_recs);

for (i=0; i< QiCDF->n_vars; i++){
  var = QiCDF->vardata[i];
  printf("Variable %d is |%s|\n", i, var->name);
  printf("sizeofentry = %ld, variable number = %ld, datatype = %ld\n",
            var->sizeofentry, var->number, var->data_type);
  printf("rec_vary = %ld, max_rec_num = %ld, num_dims = %ld, num_elems = %ld\n",
          var->rec_vary, var->max_rec_num, var->num_dims, var->num_elems); 
  printf("Dimension info:\n");
  n_total = 1;
  for (dim=0; dim < var->num_dims; dim++){
    n_total *= var->dim_sizes[dim];
    printf("Dim %d:  dim_size = %ld, dim_vary = %ld, dim_depend = %s\n",
          dim, var->dim_sizes[dim], var->dim_varys[dim], var->dim_depends[dim]);
  }
  printf("Attribute info: There are %ld variable attrs\n", var->num_v_attrs);
  for (att = 0; att < var->num_v_attrs; att++){
    printf("Attr %d has number %ld and name |%s|\n",att, 
      var->attribute[att].number, var->attribute[att].name);
    printf("data_type = %ld with %ld elements\n",var->attribute[att].data_type,
       var->attribute[att].num_elems);
    if (var->attribute->data_type == CDF_CHAR){
      printf ("CDF_CHAR attribute found: |%s|\n",(char *)var->attribute[att].data);
    }
  }
  switch (var->data_type){
     case CDF_CHAR:
       for (ival = 0; ival < n_total; ival++){
          strncpy(temp_str, var->data + ival*var->num_elems, var->num_elems);
	  *(temp_str+var->num_elems) = '\0';
          printf ("ival = %d, n_elem chars are |%s|\n",ival,temp_str);
       }
       break;
     case  CDF_FLOAT:
       for (ival = 0; ival < n_total; ival ++){
          printf("ival = %d, float data[ival] = %g\n", ival, 
	    ((float *)var->data)[ival]);
       }
       break;
     case  CDF_DOUBLE:
       for (ival = 0; ival < n_total; ival ++){
          printf("ival = %d, double data[ival] = %g\n", ival, 
	    ((double *)var->data)[ival]);
       }
       break;
     default:
       printf("Variable data type not char, float or double\n");
       break;
  } 
         
  printf("\n------------------------------\n");
}
return;

}
