/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#define CDF 0
#define QFT 1
#define QFD 2
#define TRY 3
#define CEF 4


#include "qie.h"

int main( int argc, char *argv[] )
{

QiCDFContents *QiSio;
QiOptions *QiSoptR;
QiOptions *QiSoptW;

long check = QMW_OK;
char *file_name;
char *file_name_copy;
char * header = NULL;
char * out_name = NULL;
int in_type = TRY;
int out_type;
QiOptions_priority choice = WARN;
QiOptions_rec_num rec_numbering = NUM_OFF;
char *flag;
char * ptr;
char dbg_level = 'n';
 

 
 if(argc<2)
{
  printf("Usage: %s  [-o{tdce}] [-i{tdce}] [-hName] [-nName] [-f] [-r] [-v] FileName\n",argv[0]);
  printf("options are...\n");
  printf(" -i{tdce} for input type qft, qfd, cdf, cef respectively\n");
  printf(" -o{tdce} to specify output type as  qft, qfd, cdf, cef respectively\n");
  printf(" -f forces overwrite, otherwise error returned if file exists\n");
  printf(" -r turns on record numbering if output file is flat\n");
  printf(" -v turns on  verbose messages\n");
  printf(" -hName specifies Name as that of a detached header\n");
  printf(" -nName specifies Name as that of the output file stem\n");
  printf("\n");
  printf(" Option flags MUST be space separated,\n");
  printf(" File type option must immediately follow -i and -o flags\n");
  printf(" Option -h is ignored for cdf type input file\n");
  printf(" If the input file has no extension use -i flag to set input type\n");
  exit(1);
}

argc--;
file_name = QiNewStr(argv[argc]);

if(file_name[0] == '-'){
  printf("Option flags must precede input data file\n");
  exit(1);
}

/* remove extension if present and set file type accordingly */

if((ptr=strstr(file_name,".cdf")) != NULL) {
  in_type = CDF;
  out_type = CEF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".CDF")) != NULL) {
  in_type = CDF;
  out_type = CEF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".qft")) != NULL) {
  in_type = QFT;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".QFT")) != NULL) {
  in_type = QFT;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".qfd")) != NULL) {
  in_type = QFD;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".QFD")) != NULL) {
  in_type = QFD;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".CEF")) != NULL) {
  in_type = CEF;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}
else if((ptr=strstr(file_name,".cef")) != NULL) {
  in_type = CEF;
  out_type = CDF;
  ptr[0] = '\0'; /* remove it */
}

/* if use flag to override file type, or failing that fall through to Flat TRY */

while(--argc > 0){
  flag = QiNewStr(argv[argc]);
  if(flag[0] != '-') fprintf(stderr,"unrecognised flag %s\n",flag);
  else{
    switch(flag[1]){
      case 'o':
        if      (flag[2] == 'c') out_type = CDF;
        else if (flag[2] == 't') out_type = QFT;
        else if (flag[2] == 'd') out_type = QFD;
        else if (flag[2] == 'e') out_type = CEF;
        break;
        
      case 'i':
        if      (flag[2] == 'c') in_type = CDF;
        else if (flag[2] == 't') in_type = QFT;
        else if (flag[2] == 'd') in_type = QFD;
        else if (flag[2] == 'e') in_type = CEF;
        break; 

      case 'h':
        header = QiNewStr(&(flag[2]));           
        break; 

      case 'n':
        out_name = QiNewStr(&(flag[2]));           
        break; 
            
      case 'f':
        choice = REPLACE;
        break;
        
      case 'v':
        dbg_level = 'f';
        break;
        
      case 'r':
        rec_numbering = NUM_ON;
        break;
        
      default:
       fprintf(stderr,"unrecognised flag %s\n",flag);
    }  
  }
  free(flag);
}


 /*  READ in file  */
 /*---------------------*/
 
 /* set up read options */
 
  QiSoptR = (QiOptions *) QiMakeOptionsObj();
  
  QiSoptR->fp_display = stderr;
  QiSoptR->debug_choice = dbg_level;     
  if(header == NULL) QiSoptR->header = ATTACHED;
  else {
    QiSoptR->header = DETACHED;
    QiSoptR->header_name = QiNewStr(header);
  }

  /* Initialise object to hold whole file */
 
  QiSio = (QiCDFContents *) QiMakeCDFContentsObj();  
  /* specify file name and dir */
  
  file_name_copy = QiNewStr(&(file_name[0]));
  ptr = strrchr(file_name_copy, '/');
  if(ptr != NULL){
    QiSio->io_f_name = QiNewStr(&(ptr[1]));
    ptr[1] = '\0';
    QiSio->io_f_path = QiNewStr(file_name_copy);
    QiSoptR->header_path = QiNewStr(file_name_copy);
  }
  else{
    QiSio->io_f_name = QiNewStr(file_name_copy);
  }


  switch (in_type){
    case CDF:
      check = QiGetCSDSgenCDF( &(file_name[0]), QiSio ); /* file_name includes path */
      break;
    case QFT: 
      QiSoptR->f_type = TABULAR;           
      QiSio->io_f_extn = QiSoptR->EXTN_QFT;
      check = QiGetCSDSgenFlat( QiSio, QiSoptR);
      break;
    case QFD:
      QiSoptR->f_type = DELIMITED;           
      QiSio->io_f_extn = QiSoptR->EXTN_QFD;
      check = QiGetCSDSgenFlat( QiSio, QiSoptR);
      break;
    case CEF:
      QiSoptR->f_type = EXCHANGE;           
      QiSio->io_f_extn = QiSoptR->EXTN_CEF;
      check = QiGetCSDSgenFlat( QiSio, QiSoptR);
      break;
    case TRY:
      printf("Input file type unknown, trying to read from context\n");
      QiSoptR->f_type = UNSET;           
      check = QiGetCSDSgenFlat( QiSio, QiSoptR);
      break;
    default:
     printf("unrecognised input file\n");
     exit (1);
  } /* end switch on read */
  
  if (check != QMW_OK) { 
    printf( QiErrStr(check) );
    exit(1);
  }



 /***** write new file out, same stem name *****/
 
  
 /* set up write options */

  QiSoptW = (QiOptions *) QiMakeOptionsObj();
 
  /* defaults are set up already for these options, */
  /* this provides a list of what can be set. */
  
  QiSoptW->fp_display = stderr;
  QiSoptW->debug_choice = 'f';     /* 'w'  uses ARH fn to redirect to window */
  QiSoptW->header = ATTACHED;      /* DETACHED or NO_HEADER allowed */
  QiSoptW->priority = choice;
  QiSoptW->rec_numbering = rec_numbering;
  QiSoptW->attr_delim = ',';
  QiSoptW->data_delim = ',';
  QiSoptW->rec_end = '$';
  QiSoptW->row_end = '\n';
     
 /* set output stem if changed */
  
 if(out_name != NULL) QiSio->io_f_name = QiNewStr(out_name);

 switch(out_type){
   case CDF:
     QiSio->io_f_extn = QiSoptR->EXTN_CDF;
     check = QiWriteCSDSgenCDF(QiSio, QiSoptW) ;   
     break;
   case QFT:
     QiSio->io_f_extn = QiSoptR->EXTN_QFT;
     QiSoptW->f_type = TABULAR;           
     check = QiWriteCSDSgenFlat(QiSio, QiSoptW);   
     break;
   
   case QFD:
     QiSio->io_f_extn = QiSoptR->EXTN_QFD;
     QiSoptW->f_type = DELIMITED;           
     check = QiWriteCSDSgenFlat(QiSio, QiSoptW);   
     break;

   case CEF:
     QiSio->io_f_extn = QiSoptR->EXTN_CEF;
     QiSoptW->f_type = EXCHANGE;           
     check = QiWriteCSDSgenFlat(QiSio, QiSoptW);   
     break;

   default:
     printf("cannot determine output file type\n");
     exit (1);     
 
 }  /* end switch on write */
 
 if (check != QMW_OK) { 
    printf( QiErrStr(check) );
    exit(1);
 }

  printf( "Done\n" );
  exit(0);
                
}        /* end: main */


/* redirects QSAS alert box call for stand alone */

 void QieAlertBox (char * title, char * text){
    fprintf(stderr, "%s\n", text);
    return;

}

char * QieVersion(){
 static char str[80];
  strcpy(str, QIE_VERSION); 
  return &(str[0]);
}
