/* PEACE PAD Routines
*
*  sjs
*  21 Feb 2001; 
*   6 March 2001 forward idfs files after get/fill to be ready for next spin
*  10 March 2001 working on Baz
*  11 April 2001 moved pad initialisation from pad_b_study.c to replace
*                first part (most of) GetPeacePAD
*     May   2001 Re-organised Qpeace routines
*     Oct   2001 Add Liouville Map routines
*     Jan   2002 Add promote; incomplete as new IDFS with libdbSQL... needed
*                This affects dbInitialise here, in Qsweep_spin, and includes
*                in Qidfs_utils.h
*     June 2002  Re-order so fill_pad fills using FLOW directions; half bins
*                          at ends of pitch angle range; mimics SPINPAD
*     Mar  2004  Duplicate Promote loop to update both Database and Data
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "Qpeace_pad.h"



/******************************************************************/
SDDAS_SHORT GetPeacePAD(IDFS_LINEAGE *lngpadp, IDFS_TIME *btimep,
   BIN_UNITS bin_units, PSD_UNITS psd_units, 
   T_OR_F do_pa, PEACE_SWEEP *swpp){
/******************************************************************/
/*  Get a complete PAD of PEACE data. Put it in a PEACE_SWEEP structure
*   
*   This module is self-contained
*   and deals with all IDFS calls. It needs to be passed the pointer
*   to a declared sweep structure, but it mallocs all the necessary
*   internal storage required. 
*
*   Inputs are self-evident from their declarations
*   Output is filled sweep structure
*/

void *idf_creator_pad, *idf_creator_iel, *promote_struct;
struct idf_data *idf_padp, *idf_ielp;
SDDAS_ULONG dkey_pad, dkey_iel, pa_dkey;
SDDAS_SHORT ret_val, last_sen, pa_status;
IDFS_LINEAGE lineageiel, *lngielp;
IDFS_TIME etime, btime, padsuntime, ielsuntime;
PARALLEL_FIELD_HALF_SPIN b_para_half_spin;
int b_polbin_iel,isen, isen_expected;
float b_az;
int ipromote;

lngielp = &lineageiel;
copy_lineage(lngielp,lngpadp);
strcpy(lngielp->instrument,"CORE");
strcpy(lngielp->virtual_inst,"CPCIEL");
btime = *btimep;

if (strcmp(lngpadp->instrument,"PAD")){
   printf("PAD expected, but Instr string not PAD\n");
   return PEACE_PAD_ERROR;
}

if(DIAG_PAD){
  printf("PAD bin units requested: %d <=> %s\n", 
             bin_units,BIN_UNITS_STRING(bin_units));
  printf("PAD psd units requested: %d <=> %s\n", 
             psd_units,PSD_UNITS_STRING(psd_units));
	     
}

dbInitialize ();  /* new version of IDFS needed with libdbSQL...*/
CfgInit ();
init_idfs();
get_version_number (&(lngpadp->vsn));
get_version_number (&(lngielp->vsn));

ret_val = create_idf_data_structure(&idf_creator_pad);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct for PAD\n", ret_val);
     free_experiment_info(); return ret_val;
}
idf_padp = (struct idf_data *)idf_creator_pad;

ret_val = get_data_key(lngpadp->project, lngpadp->mission,
                lngpadp->experiment, lngpadp->instrument,
                lngpadp->virtual_inst, &dkey_pad);
if (ret_val != ALL_OKAY){
       printf("Error getting data key for PAD\n");
       free_experiment_info(); return ret_val; 
}

ret_val = create_idf_data_structure(&idf_creator_iel);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct for IEL\n", ret_val);
     free_experiment_info(); return ret_val;
}
idf_ielp = (struct idf_data *)idf_creator_iel;

ret_val = get_data_key(lngielp->project, lngielp->mission,
                lngielp->experiment, lngielp->instrument,
                lngielp->virtual_inst, &dkey_iel);
if (ret_val != ALL_OKAY){
       printf("Error getting data key for CPCIEL\n");
       free_experiment_info(); return ret_val;
}
/* kludge here, because pitch_angle_source_status requires parent file_open,
*  so can't easily then do promote */
if(do_pa){
   ret_val = get_data_key(lngpadp->project, lngpadp->mission,
                   "FGM", "VEC","CFUNIT", &pa_dkey);
   if (ret_val != ALL_OKAY){
	  printf("Error getting pa_data key\n");
	  free_experiment_info();
	  return PEACE_PAD_ERROR;
   }
}

etime=btime;
etime.sec += 1;

/* do a promote if needed; do twice to try to update both database and data */

for (ipromote=0; ipromote < 2; ipromote++){  
  promote_struct = pro_InitPromote (
             btime.year, btime.doy, 1000*btime.sec, 
	     etime.year, etime.doy, 1000*etime.sec, SHOW_PROMOTE_STATUS_XWIN);
  pro_AddSource (promote_struct, dkey_pad);
  pro_AddSource (promote_struct, dkey_iel);
  /* kludge here, because pitch_angle_source_status requires parent file_open,
  *  so can't easily do promote afterwards. Only other option I think is
  *  to read pa_inst lineage directly from VIDF; this would be more generic.*/
  if(do_pa){  pro_AddSource (promote_struct, pa_dkey);}

  if (pro_NeedToPromote (promote_struct) == sTrue) {
      printf("Trying to promote data: Pass %d\n",ipromote);
      pro_PromoteAll (promote_struct);
  }
  pro_FreePromote (promote_struct);
}

ret_val = init_idf_file (btime,etime, dkey_pad, 
                                  idf_padp, lngpadp->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file for PAD\n");
   free_experiment_info(); return ret_val;
}

if(do_pa){  /* check status for pa data if requested; must do after file_open */
   ret_val = pitch_angle_source_status(
           dkey_pad,lngpadp->exten,lngpadp->vsn,&pa_status, &pa_dkey);
   if ( ret_val!= ALL_OKAY || pa_status != PA_READY ){
	printf("Pitch angle source error: return val = %d; pa_status = %d",
        	ret_val, pa_status);
	printf(" Ignoring pitch angle request\n");
	do_pa = QPFALSE;
   }
} 

/* determine last sensor for pad. This one is always filled */
ret_val = read_idf (dkey_pad, lngpadp->exten, lngpadp->vsn,
              (SDDAS_CHAR *) &last_sen, _SEN, 0, 0, 1);
if (ret_val < 0) {
      printf ("\n Error %d from read_idf routine.\n", ret_val);
} else {
   last_sen--;   /* pad always fills last sensor, which is num_sen -1 */
}     

/* read a pad record, without forwarding, to put initial vals into idf struct*/
ret_val = read_drec(dkey_pad, lngpadp->exten, lngpadp->vsn,
                     idf_padp,last_sen, 0, 1);
if (ret_val != ALL_OKAY){
   printf("1st pad read_drec failed with %d\n",ret_val);
   free_experiment_info(); return ret_val;
}                                                                          

ret_val = init_idf_file (btime, etime, dkey_iel, 
                                  idf_ielp, lngielp->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file for CPCIEL\n");
   free_experiment_info(); return ret_val;
}

/* read a iel record, without forwarding, to put initial vals into idf struct*/
ret_val = read_drec(dkey_iel, lngielp->exten, lngielp->vsn,
                     idf_ielp,14, 0, 1); /* sen 14 is b_az */
if (ret_val != ALL_OKAY){
   printf("1st pad read_drec failed with %d\n",ret_val);
   free_experiment_info(); return ret_val;
}                                                                          
 
/*  Note PAD data is time-tagged down to the sweep
*  level whereas IEL data btime/etime encompass a whole spin */

/* set pad and iel to start of (next) spin to be sure at file 
*  is at spin START 
*  If last read forwded file, this will return immediately. */

ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,lngpadp,&btime,
               &etime, last_sen);
if (ret_val != ALL_OKAY){
    printf("Failed to advance pad to a spin start\n");
    free_experiment_info(); return ret_val;
}

ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,lngielp,&btime,
               &etime, 14);
if (ret_val != ALL_OKAY){
    printf("Failed to advance iel to a spin start\n");
    free_experiment_info(); return ret_val;
}

while( (idf_ielp->sun_sen != idf_padp->sun_sen) && (ret_val == ALL_OKAY) ){
    set_times_from_idf(&btime, &etime, &padsuntime, idf_padp);
    set_times_from_idf(&btime, &etime, &ielsuntime, idf_ielp);
    switch( compare_idfs_time(&padsuntime, &ielsuntime)){
        case T1_LATER:
	  ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,
	                            lngielp,&btime, &etime, 14);
          if(ret_val != ALL_OKAY){
	       printf("Advance iel failed returning %d\n",ret_val);
	  }
	  break;
	case T2_LATER:
	  ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,
	                lngpadp,&btime, &etime, last_sen);
          if(ret_val != ALL_OKAY){
	       printf("Advance pad failed returning %d\n",ret_val);
	  }
	  break;
	case SAME_TIME: default:
	  break;
      }

}

if (ret_val != ALL_OKAY){ free_experiment_info(); return ret_val;}

if(DIAG_PAD){
printf("PAD and IEL start and end times set to spin start:\n");
set_times_from_idf(&btime,&etime,&padsuntime,idf_padp);
printf("PAD:      "); print_idfs_time(&btime); printf("\n");
printf("PAD:      "); print_idfs_time(&etime); printf("\n");
printf("PAD SUN: "); print_idfs_time(&padsuntime); printf("\n");
set_times_from_idf(&btime,&etime,&ielsuntime,idf_ielp);
printf("IEL:     "); print_idfs_time(&btime); printf("\n");
printf("IEL:     "); print_idfs_time(&etime); printf("\n");
printf("IEL SUN: "); print_idfs_time(&ielsuntime); printf("\n");
printf("IEL will be advanced prior to get_b_az call\n");
printf("PAD will be advanced after current pad is filled\n");
}

/* Note for future: to get several pads, add loop starting here.
*  See pad_b_study.c */

if(idf_padp->sun_sen != idf_ielp->sun_sen){
     printf("PAD and IEL Sun Sensor Times not same\n");
     printf("PAD: %ld   IEL %ld\n",idf_padp->sun_sen,idf_ielp->sun_sen);
     free_experiment_info(); return ret_val;
}

/* put sweep times in swpp structure */
set_times_from_idf(&(swpp->btime), &(swpp->etime), &padsuntime, idf_padp);

ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,
	              lngielp,&btime, &etime, 14);
if(ret_val != ALL_OKAY){
     printf("Advance iel failed returning %d\n",ret_val);
     free_experiment_info(); return ret_val;
}

if(DIAG_PAD){
printf("IEL advanced for get_b_az to\n");
set_times_from_idf(&btime,&etime,&ielsuntime,idf_ielp);
printf("IEL:     "); print_idfs_time(&btime); printf("\n");
printf("IEL:     "); print_idfs_time(&etime); printf("\n");
printf("IEL SUN: "); print_idfs_time(&ielsuntime); printf("\n");
}

ret_val = get_b_az(dkey_iel, idf_ielp, lngielp, &b_az, 
              &b_polbin_iel);
if(ret_val != ALL_OKAY){
     printf("Error get_b_az from iel\n");
     free_experiment_info(); return ret_val;
}

/* set sweep start, end az, n_sen, n_lev */
ret_val = init_pad(swpp,lngpadp,idf_padp,dkey_pad,
         b_az,&b_para_half_spin, &isen);
if(ret_val!=ALL_OKAY){
     printf("Error init_pad\n");
     free_experiment_info(); return ret_val;
}

isen_expected = expected_first_polbin(b_polbin_iel,lngpadp,swpp);

if (isen != isen_expected){
printf("isen_expected = %d differs from empirical one",isen_expected);
printf(" %d. Using empirical\n", isen);
}

if (DIAG_PAD){
printf("%d energy levels per sweep; %d pitch angles\n",
            swpp->n_lev, swpp->n_sen);
printf("b_polbin_iel, b_az(iel), isen(1st half) b_para_half_spin, ");
printf("isen_expected\n");
printf("       %d      %g           %d              %d            %d\n",
   b_polbin_iel, b_az, isen, b_para_half_spin, isen_expected);
}

ret_val = malloc_sweep_arrays(swpp, do_pa);
if (ret_val != ALL_OKAY){
     printf("malloc_sweep_arrays failed\n");
     free_experiment_info(); return ret_val;
}
swpp->missing_value = QP_DEFAULT_MISSING;

ret_val = e_bin_edges(dkey_pad, idf_padp, lngpadp->vsn, swpp->elo,
                 swpp->ehi, bin_units, swpp->n_lev);
if (ret_val != ALL_OKAY){
   printf("Error getting E bin edges returned %d\n",ret_val);
   free_sweep_arrays(swpp); free_experiment_info(); return ret_val;
}

ret_val = invert_array(swpp->elo,swpp->n_lev);
if (ret_val == ALL_OKAY){
  ret_val = invert_array(swpp->ehi,swpp->n_lev);
  if (ret_val == ALL_OKAY){
     ret_val = check_bins(swpp->elo, swpp->ehi, swpp->n_lev);
  }
}
if (ret_val != ALL_OKAY){
   printf("Error inverting or checking e bins\n");
   free_sweep_arrays(swpp); free_experiment_info(); return ret_val;
}

ret_val = set_equal_bins_half_ends(swpp->thlo,swpp->thhi,swpp->n_sen,0.0,180.);
if (ret_val != ALL_OKAY){
   printf("Error setting polar bins\n");
   free_sweep_arrays(swpp); free_experiment_info(); return ret_val;
}

copy_lineage(swpp->lineage, lngpadp);

if(DIAG_PAD){
   printf("About to fill pad. idf_padp structure is \n");
   print_idf_data(idf_padp);
}

ret_val = fill_pad(swpp,lngpadp,idf_padp,dkey_pad,
                 bin_units,psd_units,do_pa, b_para_half_spin,isen);
if (ret_val != ALL_OKAY){
   printf("Error fill_pad returned %d \n",ret_val);
   free_sweep_arrays(swpp); free_experiment_info(); 
   printf("fill_pad returning\n"); return ret_val;
}
free_experiment_info();
return ret_val;

}

/******************************************************************/
SDDAS_SHORT GetNPeacePADs(
        IDFS_LINEAGE *lngpadp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *npads, 
	PEACE_SWEEP ***swpppp){
/******************************************************************/
/*  Get all PEACE PAD data within times specified. 
*   Put them each in a PEACE_SWEEP structure and return an array of
*   pointers which points to them.
*   
*   This module is self-contained
*   and deals with all IDFS calls. It returns the pointer
*   to an array of sweep structure pointers, and mallocs all the necessary
*   internal storage required, including the pointer array. 
*
*   Inputs are self-evident from their declarations
*   Output is malloced and filled pointer array (and the npads, sweep 
*   structures to which they point!). Sweep times altered to spin start/stop
*/

void *idf_creator_pad, *idf_creator_iel, *promote_struct;
struct idf_data *idf_padp, *idf_ielp;
SDDAS_ULONG dkey_pad, dkey_iel, pa_dkey;
SDDAS_SHORT ret_val, last_sen, pa_status;
IDFS_LINEAGE lineageiel, *lngielp;
IDFS_TIME etime, btime, lasttime, padsuntime, ielsuntime, dummytime;
PARALLEL_FIELD_HALF_SPIN b_para_half_spin;
int b_polbin_iel,isen, isen_expected;
float b_az;
PEACE_SWEEP *curswp, **swppp;
int i, ipad, np;   /* np is local copy of *npads */
int ipromote;

lngielp = &lineageiel;
copy_lineage(lngielp,lngpadp);
strcpy(lngielp->instrument,"CORE");
strcpy(lngielp->virtual_inst,"CPCIEL");
btime = *btimep;

if (strcmp(lngpadp->instrument,"PAD")){
   printf("PAD expected, but Instr string not PAD\n");
   return PEACE_PAD_ERROR;
}

if(DIAG_PAD){
  printf("PAD bin units requested: %d <=> %s\n", 
             bin_units,BIN_UNITS_STRING(bin_units));
  printf("PAD psd units requested: %d <=> %s\n", 
             psd_units,PSD_UNITS_STRING(psd_units));
	     
}

dbInitialize ();  /* new version of IDFS needed with libdbSQL... */
CfgInit ();
init_idfs();
get_version_number (&(lngpadp->vsn));
get_version_number (&(lngielp->vsn));

ret_val = create_idf_data_structure(&idf_creator_pad);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct for PAD\n", ret_val);
     free_experiment_info(); return ret_val;
}
idf_padp = (struct idf_data *)idf_creator_pad;

ret_val = get_data_key(lngpadp->project, lngpadp->mission,
                lngpadp->experiment, lngpadp->instrument,
                lngpadp->virtual_inst, &dkey_pad);
if (ret_val != ALL_OKAY){
       printf("Error getting data key for PAD\n");
       free_experiment_info(); return ret_val; 
}

ret_val = create_idf_data_structure(&idf_creator_iel);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct for IEL\n", ret_val);
     free_experiment_info(); return ret_val;
}
idf_ielp = (struct idf_data *)idf_creator_iel;

ret_val = get_data_key(lngielp->project, lngielp->mission,
                lngielp->experiment, lngielp->instrument,
                lngielp->virtual_inst, &dkey_iel);
if (ret_val != ALL_OKAY){
       printf("Error getting data key for CPCIEL\n");
       free_experiment_info(); return ret_val;
}
/* kludge here, because pitch_angle_source_status requires parent file_open,
*  so can't easily then do promote */
if(do_pa){
   ret_val = get_data_key(lngpadp->project, lngpadp->mission,
                   "FGM", "VEC","CFUNIT", &pa_dkey);
   if (ret_val != ALL_OKAY){
	  printf("Error getting pa_data key\n");
	  free_experiment_info();
	  return PEACE_PAD_ERROR;
   }
}

etime=btime;
etime.sec += 1;

/* do a promote if needed; twice to update database and data */

for (ipromote=0; ipromote<2; ipromote++){
   promote_struct = pro_InitPromote (
              btime.year, btime.doy, 1000*btime.sec, 
	      etime.year, etime.doy, 1000*etime.sec, SHOW_PROMOTE_STATUS_XWIN);
   pro_AddSource (promote_struct, dkey_pad);
   pro_AddSource (promote_struct, dkey_iel);
   /* kludge here, because pitch_angle_source_status requires parent file_open,
   *  so can't easily do promote afterwards. Only other option I think is
   *  to read pa_inst lineage directly from VIDF; this would be more generic.*/
   if(do_pa){  pro_AddSource (promote_struct, pa_dkey);}

   if (pro_NeedToPromote (promote_struct) == sTrue) {
       printf("Trying to promote data; pass %d\n",ipromote);
       pro_PromoteAll (promote_struct);
   }
   pro_FreePromote (promote_struct);
}

ret_val = init_idf_file (btime,etime, dkey_pad, 
                                  idf_padp, lngpadp->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file for PAD\n");
   free_experiment_info(); return ret_val;
}

if(do_pa){  /* check status for pa data if requested; must do after file_open */
   ret_val = pitch_angle_source_status(
           dkey_pad,lngpadp->exten,lngpadp->vsn,&pa_status, &pa_dkey);
   if ( ret_val!= ALL_OKAY && pa_status != PA_READY ){
        printf("Pitch angle source error: return val = %d; pa_status = %d",
        	ret_val, pa_status);
	printf(" Ignoring pitch angle request\n");
	do_pa = QPFALSE;
   }
} 


/* determine last sensor for pad. This one is always filled */
ret_val = read_idf (dkey_pad, lngpadp->exten, lngpadp->vsn,
              (SDDAS_CHAR *) &last_sen, _SEN, 0, 0, 1);
if (ret_val < 0) {
      printf ("\n Error %d from read_idf routine.\n", ret_val);
} else {
   last_sen--;   /* pad always fills last sensor, which is num_sen -1 */
}     

/* read a pad record, without forwarding, to put initial vals into idf struct*/
ret_val = read_drec(dkey_pad, lngpadp->exten, lngpadp->vsn,
                     idf_padp,last_sen, 0, 1);
if (ret_val != ALL_OKAY){
   printf("1st pad read_drec failed with %d\n",ret_val);
   free_experiment_info(); return ret_val;
}                                                                          

ret_val = init_idf_file (btime, etime, dkey_iel, 
                                  idf_ielp, lngielp->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file for CPCIEL\n");
   free_experiment_info(); return ret_val;
}

/* read a iel record, without forwarding, to put initial vals into idf struct*/
ret_val = read_drec(dkey_iel, lngielp->exten, lngielp->vsn,
                     idf_ielp,14, 0, 1); /* sen 14 is b_az */
if (ret_val != ALL_OKAY){
   printf("1st pad read_drec failed with %d\n",ret_val);
   free_experiment_info(); return ret_val;
}                                                                          
 
/*  Note PAD data is time-tagged down to the sweep
*  level whereas IEL data btime/etime encompass a whole spin */

/* set pad and iel to start of (next) spin to be sure at file 
*  is at spin START 
*  If last read forwded file, this will return immediately. */

ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,lngpadp,&btime,
               &etime, last_sen);
if (ret_val != ALL_OKAY){
    printf("Failed to advance pad to a spin start\n");
    free_experiment_info(); return ret_val;
}

ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,lngielp,&btime,
               &etime, 14);
if (ret_val != ALL_OKAY){
    printf("Failed to advance iel to a spin start\n");
    free_experiment_info(); return ret_val;
}

while( (idf_ielp->sun_sen != idf_padp->sun_sen) && (ret_val == ALL_OKAY) ){
    set_times_from_idf(&btime, &etime, &padsuntime, idf_padp);
    set_times_from_idf(&btime, &etime, &ielsuntime, idf_ielp);
    switch( compare_idfs_time(&padsuntime, &ielsuntime)){
        case T1_LATER:
	  ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,
	                            lngielp,&btime, &etime, 14);
          if(ret_val != ALL_OKAY){
	       printf("Advance iel failed returning %d\n",ret_val);
	  }
	  break;
	case T2_LATER:
	  ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,
	                lngpadp,&btime, &etime, last_sen);
          if(ret_val != ALL_OKAY){
	       printf("Advance pad failed returning %d\n",ret_val);
	  }
	  break;
	case SAME_TIME: default:
	  break;
      }

      printf("Initial try match PAD Sun sensor = %ld; IEL Sun sensor = %ld\n",
            idf_padp->sun_sen, idf_ielp->sun_sen);
}

if (ret_val != ALL_OKAY){ free_experiment_info(); return ret_val;}

if(DIAG_PAD){
printf("PAD and IEL start and end times set to spin start:\n");
set_times_from_idf(&btime,&etime,&padsuntime,idf_padp);
printf("PAD:      "); print_idfs_time(&btime); printf("\n");
printf("PAD:      "); print_idfs_time(&etime); printf("\n");
printf("PAD SUN: "); print_idfs_time(&padsuntime); printf("\n");
set_times_from_idf(&btime,&etime,&ielsuntime,idf_ielp);
printf("IEL:     "); print_idfs_time(&btime); printf("\n");
printf("IEL:     "); print_idfs_time(&etime); printf("\n");
printf("IEL SUN: "); print_idfs_time(&ielsuntime); printf("\n");
printf("IEL will be advanced prior to get_b_az call\n");
printf("PAD will be advanced after current pad is filled\n");
}

/* get several pads See pad_b_study.c */

np = (int)((idfs_time_to_epoch_secs(etimep) - 
                idfs_time_to_epoch_secs(btimep))/(idf_padp->spin_rate/1000.));
np++;  /* expect at most np pads */

if(DIAG_PAD){printf("Expecting at most %d spins\n",np);}
if (np < 1){
     printf("GetNPeacePADs npads = %d out of range\n", np);
     free_experiment_info(); return PEACE_PAD_ERROR;
}
swppp= (PEACE_SWEEP**)malloc(np*sizeof(PEACE_SWEEP*));
if (swppp == NULL){
     printf("GetNPeacePADs failed to malloc swppp with %d sweep ptrs\n",np);
     free_experiment_info();  return PEACE_PAD_ERROR;
}

*swpppp = swppp;  /* pass the address of the ptr array back */

for (i=0; i<np; swppp[i++]=NULL);  /* initialise ptrs to NULL */

btime = *btimep;
lasttime = *etimep;
ipad = 0;

while ((compare_idfs_time(&btime,&lasttime)!=T1_LATER )&&
     (ret_val == ALL_OKAY && ipad < np)){

     if(idf_padp->sun_sen != idf_ielp->sun_sen){
	  printf("PAD and IEL Sun Sensor Times not same\n");
	  printf("PAD: %ld   IEL %ld\n",idf_padp->sun_sen,idf_ielp->sun_sen);
	      set_times_from_idf(&dummytime, &dummytime, &padsuntime, idf_padp);
              set_times_from_idf(&dummytime, &dummytime, &ielsuntime, idf_ielp);
              switch( compare_idfs_time(&padsuntime, &ielsuntime)){
                 case T1_LATER:
	            ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,
	                            lngielp,&btime, &etime, 14);
                    if(ret_val != ALL_OKAY){
	              printf("Advance iel failed returning %d\n",ret_val);
	            }
	            break;
	         case T2_LATER:
	            ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,
	                lngpadp,&btime, &etime, last_sen);
                    if(ret_val != ALL_OKAY){
	               printf("Advance pad failed returning %d\n",ret_val);
	            }
	            break;
	            case SAME_TIME: default:
	            break;
               }
	  printf("PAD and IEL Sun Sensor are now\n");
	  printf("PAD: %ld   IEL %ld\n",idf_padp->sun_sen,idf_ielp->sun_sen);
	  continue;
     }

     curswp = swppp[ipad] = (PEACE_SWEEP*)malloc(sizeof(PEACE_SWEEP));
     if (curswp == NULL){
        printf("GetNPads failed to malloc sweep structure\n");
	break;
     }
     nullify_sweep_arrays(curswp);  /* set ptrs inside swp to NULL for safety*/
     /* put SPIN start time as sweep start in swpp structure - for continuous
     *  plot; end time will come from iel after advance!  */
     set_times_from_idf(&(dummytime), &(dummytime), 
                             &padsuntime, idf_padp);
     curswp->btime = padsuntime;

     ret_val = advance_idf_file_spin_start(dkey_iel,idf_ielp,
	        	   lngielp,&btime, &etime, 14);
     if(ret_val != ALL_OKAY){
	  printf("Advance iel failed returning %d\n",ret_val);
	  break;
     }

     set_times_from_idf(&btime,&etime,&ielsuntime,idf_ielp);
     curswp->etime = ielsuntime;
     
     if(DIAG_PAD){
     printf("IEL advanced for get_b_az to\n");
     printf("IEL:     "); print_idfs_time(&btime); printf("\n");
     printf("IEL:     "); print_idfs_time(&etime); printf("\n");
     printf("IEL SUN: "); print_idfs_time(&ielsuntime); printf("\n");
     }

     ret_val = get_b_az(dkey_iel, idf_ielp, lngielp, &b_az, 
        	   &b_polbin_iel);
     if(ret_val != ALL_OKAY){
	  printf("Error get_b_az from iel\n");
	  break;
     }

     /* set sweep start, end az, n_sen, n_lev */
     ret_val = init_pad(curswp,lngpadp,idf_padp,dkey_pad,
                                b_az,&b_para_half_spin, &isen);
     if(ret_val!=ALL_OKAY){
	  printf("Error init_pad\n");
	  break;
     }

     isen_expected = expected_first_polbin(b_polbin_iel,lngpadp,curswp);

     if (isen != isen_expected){
     printf("isen_expected = %d differs from empirical one",isen_expected);
     printf(" %d. Using empirical\n", isen);
     }

     if (DIAG_PAD){
     printf("%d energy levels per sweep; %d pitch angles\n",
        	 curswp->n_lev, curswp->n_sen);
     printf("b_polbin_iel, b_az(iel), isen(1st half) b_para_half_spin, ");
     printf("isen_expected\n");
     printf("       %d      %g           %d              %d            %d\n",
	b_polbin_iel, b_az, isen, b_para_half_spin, isen_expected);
     }

     ret_val = malloc_sweep_arrays(curswp, do_pa);
     if (ret_val != ALL_OKAY){
	  printf("malloc_sweep_arrays failed\n");
	  break;
     }
     curswp->missing_value = QP_DEFAULT_MISSING;
     
     ret_val = e_bin_edges(dkey_pad, idf_padp, lngpadp->vsn, curswp->elo,
                      curswp->ehi, bin_units, curswp->n_lev);
     if (ret_val != ALL_OKAY){
	printf("Error getting E bin edges returned %d\n",ret_val);
	break;  /* ?? free_sweep_arrays */
     }

     ret_val = invert_array(curswp->elo,curswp->n_lev);
     if (ret_val == ALL_OKAY){
       ret_val = invert_array(curswp->ehi,curswp->n_lev);
       if (ret_val == ALL_OKAY){
	  ret_val = check_bins(curswp->elo, curswp->ehi, curswp->n_lev);
       }
     }
     if (ret_val != ALL_OKAY){
	printf("Error inverting or checking e bins\n");
	break; /* ??? */
     }

     ret_val = set_equal_bins_half_ends(curswp->thlo,curswp->thhi,
                                            curswp->n_sen,0.0,180.);
     if (ret_val != ALL_OKAY){
	printf("Error setting polar bins\n");
	break;
     }

     copy_lineage(curswp->lineage,lngpadp);

     if(DIAG_PAD){
	printf("About to fill pad. idf_padp structure is \n");
	print_idf_data(idf_padp);
     }

     ret_val = fill_pad(curswp,lngpadp,idf_padp,dkey_pad,
                      bin_units,psd_units,do_pa, b_para_half_spin,isen);
     if (ret_val != ALL_OKAY){
	printf("Error fill_pad returned %d \n",ret_val); 
	printf("fill_pad returning\n"); break;
     }

     /* advance pad file to next spin */
     ret_val = advance_idf_file_spin_start(dkey_pad,idf_padp,
	        	   lngpadp,&btime, &etime, last_sen);
     if(ret_val != ALL_OKAY){
      printf("Advance pad failed returning %d\n",ret_val);
      break;
     }

     if(DIAG_PAD){
     printf("PAD advanced after fill_pad to\n");
     set_times_from_idf(&btime,&etime,&padsuntime,idf_padp);
     printf("PAD:     "); print_idfs_time(&btime); printf("\n");
     printf("PAD:     "); print_idfs_time(&etime); printf("\n");
     printf("PAD SUN: "); print_idfs_time(&padsuntime); printf("\n");
     }

     ipad++; 
     
}  /* end while loop gathering pads */


if (ret_val != ALL_OKAY || ipad <= 0 || ipad > np){
    printf("GetNPads Error after gathering while: Returned %d with ipad =%d",
                ret_val, ipad);
    printf(" out of %d expected\n", np);
    printf("Current spin start time = ");
    print_idfs_time(&btime);printf("\n");
    if (ret_val == NEXT_FILE_STATUS && ipad > 0 && ipad <=np){
        printf("NEXT_FILE_STATUS reached\n");
    } else {  
        free_n_sweeps(np, swppp);
        return PEACE_PAD_ERROR;
    }
}
*npads = ipad;  /* set to actual number of pads filled; this is
                   also the number of sweep structures malloced, 
		   although the size of the pointer array could be
		   bigger */

free_experiment_info();
return ret_val;

}

/*******************************************************************************/
SDDAS_SHORT init_pad(PEACE_SWEEP *swpp, IDFS_LINEAGE *lngp, 
    struct idf_data *idfp, SDDAS_ULONG dkey,
    float b_az, PARALLEL_FIELD_HALF_SPIN *b_para_half_spin,
    int *isen_init){
/*******************************************************************************/
/* get some info about pad instrument from vidf and by finding a filled data record.
*  File should be positioned prior to call
*  Info returned in swpp includes:
*
*  n_sen   = number of sensors  [should <= MAX_P_ANGLES -1]
*  n_sam   = number of energy levels in sweep
*  lineage
*  phi_lo, phi_hi  - based on b_az and sweep width
*
*  Info for pad sorting:
*  b_para_half_spin = flag for whether field_aligned bin is in 1st or 2nd half spin
*/

SDDAS_SHORT rval, ret_val;
SDDAS_SHORT num_sensor;
SDDAS_CHAR fwd, full_swp;
int isen;

rval = read_idf (dkey, lngp->exten, lngp->vsn,
              (SDDAS_CHAR *) &num_sensor, _SEN, 0, 0, 1);
if (rval < 0) {
      printf ("\n Error %d from read_idf routine.\n", rval);
      return PEACE_PAD_ERROR;
    }
if (num_sensor > MAX_P_ANGLES -1){
      printf ("Pad has too many %d sensors! Oops.\n",MAX_P_ANGLES);
      return PEACE_PAD_ERROR;
}
swpp->n_sen = num_sensor+1;  /* number of pitch angles */

fwd = 0; full_swp = 1;isen = 0; /* read until find filled sensor; don't advance t */
idfp->filled_data = 0;   /* trick into 1st read */
while (!idfp->filled_data && isen < num_sensor){
     rval = read_drec (dkey, lngp->exten, lngp->vsn, idfp, isen,fwd, full_swp);
     if (rval < 0){
            printf ("\nError %d from pad read_drec.\n", rval);
            return PEACE_PAD_ERROR;
     }
     isen++;
     if(DIAG_PAD){ 
         printf("read_drec'ed sensor %d. ret_val = %d, filled_data = %d\n",
                      isen-1, rval, idfp->filled_data);
     }
}
if (isen == num_sensor  && !idfp->filled_data){
  printf("Failed to initialise; all sensors read but no data filled\n");
  return PEACE_PAD_ERROR;
}
swpp->n_lev = idfp->num_sample;
isen -= 1;         /* sensor found; counter incremented once too many times */
*isen_init = isen;

if(DIAG_PAD)printf("Initialise used sensor %d; there are %d n_samples\n",
        isen, swpp->n_lev);

/* set flag for which half spin contains the parallel field direction*/

ret_val = set_b_para_half_spin(b_az, lngp, b_para_half_spin);
if (ret_val != ALL_OKAY){
   printf("ERROR:set_b_para_half_spin returned %d\n",ret_val);
   return ret_val;
}

if(DIAG_PAD){ 
     printf("b_az = %g, b_para_half_spin set to %d (0=1st half)\n",
              b_az, *b_para_half_spin);
}
swpp->phi_frame = PEACE_TZERO;    /* not really needed, but safe */
swpp->lineage = lngp;
swpp->phi_lo = mod360(idfp->start_az[0]+180.0);
swpp->phi_hi = mod360(idfp->stop_az[idfp->num_angle-1]+180.0);

return ALL_OKAY;
}

/**********************************************************************************/
SDDAS_SHORT  fill_pad(PEACE_SWEEP *swpp,
                      IDFS_LINEAGE *lineage,
		      struct idf_data *idf_padp,
		      SDDAS_ULONG dkey_pad, 
		      BIN_UNITS bin_units,
                      PSD_UNITS psd_units,
		      T_OR_F do_pa,
		      PARALLEL_FIELD_HALF_SPIN b_half_spin,
		      int isen){
/********************************************************************************/
/* fill the psd array in a sweep structure with a PAD
*  in the appropriate units by reading both halves of spin 
*
*  File should be positioned prior to call and is forwarded when
*  last sensor is read */
/* psd array is dimensioned psd[pa_dim][level_dim]  */

SDDAS_SHORT ret_val,num_sensor;
SDDAS_LONG ret_vallong;
SDDAS_FLOAT hold_data[MAX_P_ANGLES][MAX_LEVELS];  /* to receive the data */
SDDAS_FLOAT hold_padata[MAX_P_ANGLES][MAX_LEVELS];/* to receive the pa data */
SDDAS_FLOAT mode_ret;
SDDAS_LONG  bmilli_first_half;

SDDAS_SHORT tbl_oper[MAX_TABLES_TO_APPLY];
SDDAS_CHAR num_tbls, tbls_to_apply[MAX_TABLES_TO_APPLY];
SDDAS_CHAR full_swp, fwd;
PARALLEL_FIELD_HALF_SPIN b_para_half_copy;
T_OR_F harskip;

int inbins[MAX_P_ANGLES];  /* hold position round full circle */
int outbins[MAX_P_ANGLES]; /* sorted indices for copying data to pad array */
int inum_sensor,n_pa, swp_half, bin_got, b_polbin;

int ipa,jlev,jdlev, nlev;
int i,k, iret, ii;

swpp->bin_units = bin_units;
swpp->psd_units = psd_units;

if(DIAG_PAD){
  printf("PAD bin units set to: %d <=> %s\n", 
             bin_units,BIN_UNITS_STRING(bin_units));
  printf("PAD psd units set to: %d <=> %s\n", 
             psd_units,PSD_UNITS_STRING(psd_units));
	     
}
n_pa = swpp->n_sen;  /* sweep has n_sen = number of distinct sensors + 1 */
nlev = swpp->n_lev;  /* local copy */

/* get number of sensors and check against expected */
ret_vallong = read_idf (dkey_pad, lineage->exten, 
             lineage->vsn, (SDDAS_CHAR *) &num_sensor, _SEN, 0, 0, 1);
if (ret_vallong < 0) {
      printf ("\n Error %ld from read_idf routine in fill_single_pad.\n", ret_vallong);
      return PEACE_PAD_ERROR;
}
inum_sensor = num_sensor;
if (inum_sensor != n_pa-1){
      printf ("Pad has %d sensors; expected:(n_pitch_angles - 1)= %d! Oops.\n",
                             num_sensor, n_pa - 1);
      return PEACE_PAD_ERROR;
}
ret_val = set_psd_tables(lineage, psd_units, &num_tbls, tbls_to_apply, tbl_oper);
if (ret_val != ALL_OKAY){
   printf("Set_psd_tables failed\n"); return PEACE_PAD_ERROR;
}

for (jlev=0; jlev<MAX_LEVELS; jlev++){
   for (ipa=0; ipa<MAX_P_ANGLES; ipa++){
      hold_data[ipa][jlev]=-1.0;  /* fill to error check */
   }
}

if(DIAG_PRINT)printf("initial part of fill_single_pad complete\n ");

/* see if we are in HAR mode. If so, there is an extra read/foward before
*  and after we fill each half-spin in order to (a) select the middle
*  of the 3 HAR slices and (b) get to the 2nd half-spin
*
*  Oops, it would appear that HAR doesn't (always?) have 3 slices, sometimes
*  only 2. The only quick solution is to just use the time info and keep
*  advancing to the next half spin; I'll do that but leave in the harskip
*  code in case there is a way to pin-down the number of azimuthal slices
*  per half-spin in HAR
*/
/* First: re-read some data (just to be sure; init_pad should have just
*  done that), check the mode and if har set harskip to QPTRUE 
*  Also, store  bmilli_first_half */
fwd=0; full_swp=1;
ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                        idf_padp, num_sensor -1,fwd,full_swp);
if (ret_val < 0){
       printf ("\nError %d from pad read_drec. to set harskip\n", ret_val);
       return PEACE_PAD_ERROR;
}
ret_val = convert_to_units (dkey_pad, lineage->exten, lineage->vsn,
            idf_padp, 0, MODE, 0,  /*sensor used to indic which status byte*/
	    0, tbls_to_apply, tbl_oper, &mode_ret, 0, 0);
harskip = (mode_ret == 2) ? QPTRUE : QPFALSE;
if(DIAG_PAD){
   if (idf_padp->filled_data){printf("Read to set harskip filled data ok\n");
   }else{ printf("Read to set harskip filled data failed\n");}
   printf("mode_ret = %f and harskip set to %d\n",mode_ret,harskip);
}
bmilli_first_half = idf_padp->bmilli;

/* end harskip prep */

full_swp = 1;
bin_got = -1;   /* index of number of sensors <=> bin got */

for (swp_half = 0; swp_half <2; swp_half++){    /* read in two halves of spin */
  /* Don't skip at beginning for har; always grap 1st az slice
  if(harskip){   
     ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                             idf_padp, num_sensor -1, 1 , full_swp);
     if (ret_val < 0){
            printf ("\nError %d from pad read_drec. in pre-harskip\n", ret_val);
            return PEACE_PAD_ERROR;
     }
     if(DIAG_PAD){
          printf("Pre-Harskip: Read sensor %d with time %ld\n",
	  num_sensor-1, idf_padp->bmilli);
     }
  } */
  for (i=0; i< num_sensor; i++){
     if ( i == num_sensor -1) {
         fwd = 1; /* fwd to 2nd 1/2 spin - swp_half=0, or next spin */
     } else fwd = 0; 
     ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                             idf_padp, i, fwd, full_swp);
     if (ret_val < 0){
            printf ("\nError %d from pad read_drec.\n", ret_val);
            return PEACE_PAD_ERROR;
     }
     if(DIAG_PAD){
          printf("Read sensor %d, filled_data %d with time %ld\n",
	  i, idf_padp->filled_data,idf_padp->bmilli);
     }
     if (idf_padp->filled_data){   /* yes; got some */
        bin_got++;
        if (bin_got > n_pa -1){
           printf("PAD ERROR: More pa bins got than n_pa expected\n"); 
	   return PEACE_PAD_ERROR;
        }
        ret_val = convert_to_units (dkey_pad, lineage->exten, lineage->vsn, 
	         idf_padp, i, SENSOR, 0, 
		 num_tbls, tbls_to_apply, tbl_oper, hold_data[bin_got], 0, 0);
	if (ret_val != ALL_OKAY){
	   printf("Error converting to units in fill_single_pad\n"); 
	   return PEACE_PAD_ERROR;
	}
        if (do_pa){
	  ret_val = convert_to_units (dkey_pad, lineage->exten, lineage->vsn, 
	           idf_padp, i, PITCH_ANGLE, 0, 
		   0, tbls_to_apply, tbl_oper, hold_padata[bin_got], 0, 0);
	  if (ret_val != ALL_OKAY){
	     printf("Error converting to units for pa fill_single_pad\n"); 
	     return PEACE_PAD_ERROR;
	  }
	}
	if (DIAG_PAD_IDF){
	   print_idf_data(idf_padp); printf("\nPitch angles for sweep:\n");
	   for (ii =0; ii<nlev; ii++){
	      printf("level = %d, pa = %g\n",ii,hold_padata[bin_got][ii]);
	   }
	}
        inbins[bin_got] = compute_clockbin(i,num_sensor,swp_half,n_pa);
        if (inbins[bin_got] < 0){
           printf("Error from compute_clockbin\n"); return PEACE_PAD_ERROR;
        }
    }
  }
  if(harskip){   /* if HAR get to next 1/2 spin; nb last read above  will */
                 /* have forwarded, so do a read 1st */
     ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                               idf_padp, num_sensor -1, 0 , full_swp);
     while(idf_padp->bmilli - bmilli_first_half < PAD_HAR_HALF_SPIN_MILLIS){
       ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                               idf_padp, num_sensor -1, 1 , full_swp);
       if (ret_val < 0){
         printf ("\nError %d from pad read_drec. in post-harskip\n", ret_val);
         return PEACE_PAD_ERROR;
       } 
       if(DIAG_PAD){
         printf("Post-Harskip: Mid-Read sensor %d @ time %ld cf 1st half %ld\n",
	 num_sensor-1, idf_padp->bmilli, bmilli_first_half);
       }
       /* after forward; do a non-fwd read to set up new time in idfp: */
       ret_val = read_drec (dkey_pad, lineage->exten, lineage->vsn, 
                               idf_padp, num_sensor -1, 0 , full_swp);
       if (ret_val < 0){
         printf ("\nError %d from pad read_drec. in post-harskip\n", ret_val);
         return PEACE_PAD_ERROR;
       }
       if(DIAG_PAD){
         printf("Post-Harskip: Read sensor %d with time %ld cf 1st half %ld\n",
	 num_sensor-1, idf_padp->bmilli, bmilli_first_half);
       }
     }
  }
}

if (++bin_got != n_pa){
   printf("Got/filled %d pitch angle bins;  expected %d\n",bin_got,n_pa);
   return PEACE_PAD_ERROR;
}

/* set b_polbin to +B look direction given b_half_spin info on which
*  half-spin contains this direction and isen, which is the 1st bin
*  in the 1st half spin which actually contains data.    */

switch (b_half_spin){
   case PARALLEL_FIRST_HALF:   /* 1st half spin has +B, so 1st filled */
      b_polbin = isen;         /* sensor is in +B look direction */
      break;
   case PARALLEL_SECOND_HALF:  /* 1st filled sensor in 1st half is -B look */
      b_para_half_copy = PARALLEL_FIRST_HALF;
      b_polbin = isen;
      iret = exchange_b_tip_tail(&b_polbin, &b_para_half_copy, n_pa-1);
      if (iret == PEACE_PAD_ERROR) { 
         printf("Error from exchange_b_tip_tail in fill_pad switch\n");
	 return PEACE_PAD_ERROR;
      }
      break;
   default:
      printf("b_half_spin invalid in fill_pad switch\n");
      return PEACE_PAD_ERROR;
      break;
}
         
/* inbins holds clock angle of all sweeps
*  go sort them in order starting with field-aligned FLOW direction
*  outbins will contain indices of sweeps in this order */
ret_val = sort_pa_bins(inbins, outbins, n_pa, b_half_spin, b_polbin);
if (ret_val != ALL_OKAY){
   printf("PAD ERROR: sort_pa_bins returned error\n");
   return PEACE_PAD_ERROR;
}

/* copy data into sweep using outbins to index pa dimension
*  re-order energy levels from lowest to highest */
for (ipa=0; ipa<n_pa; ipa++){   /* psd indices [ang][en] */
   k = nlev * ipa;
   for (jlev=0, jdlev = nlev -1; jlev<nlev; jlev++,jdlev--){
      swpp->psd[k+jlev] = hold_data[outbins[ipa]][jdlev];                                 
   }
}
if (do_pa){
  for (ipa=0; ipa<n_pa; ipa++){   /* pa indices [ang][en] */
     k = nlev * ipa;
     for (jlev=0, jdlev = nlev -1; jlev<nlev; jlev++,jdlev--){
	swpp->pa[k+jlev] = hold_padata[outbins[ipa]][jdlev];                                 
     }
  }
}
return ALL_OKAY;
}

/************************************************************************/
int expected_first_polbin(int b_polbin_iel, IDFS_LINEAGE *lngp, 
                             PEACE_SWEEP *swpp){
/************************************************************************/
/* calculated expected 1st polar bin in 1st half of spin to be filled.
*  See notes for set_b_para_half_spin for explanation. Basically,
*  if LEEA this is b_polbin as returned in iel. If HEEA it is flipped
*  to n_sensors -1 - (b_polbin_iel)
*/

if (strcmp(lngp->virtual_inst,"CPPADL")==0){
   return b_polbin_iel;
} else if (strcmp(lngp->virtual_inst,"CPPADH")==0){
   return swpp->n_sen -2 - b_polbin_iel; 
   /* swpp holds pitch angles, so swpp->n_sen = n_sensors + 1*/
} 

printf("expected_first_polbin Virt Inst = %s, not CPPADL or CPPADH\n",
                   lngp->virtual_inst);
return b_polbin_iel;  /* not correct but safe */
}


/*************************************************************************/
SDDAS_SHORT get_b_az(SDDAS_ULONG dkey, struct idf_data *idfp, 
                      IDFS_LINEAGE *lineage, float *b_az, 
		      int *b_pol_iel){
/*************************************************************************/
/* find the azimuthal and polar info for B from iel sensor
*
********  NB must be given dkey which corresponds to iel sensor!
*
*  File should already be positioned and will be forwarded
*  (but NOT read) before return.
*/

SDDAS_CHAR num_tbls=0, tbls_to_apply[4];
SDDAS_SHORT tbl_oper[4];
SDDAS_FLOAT baz_sdd,bpol_sdd;
SDDAS_CHAR fwd=1, no_fwd=0;

SDDAS_SHORT rcode, ret_val;

/* sensor 14 is B-field azimuth in degrees */

ret_val = read_drec (dkey, lineage->exten, lineage->vsn, idfp, 14,no_fwd, 0);
if (ret_val < 0){
            printf ("\nError %d from read_drec. iel sens 14\n", ret_val);
            return PEACE_PAD_ERROR;
}
if (ret_val == ALL_OKAY || idfp->filled_data){
   rcode = convert_to_units (dkey, lineage->exten, lineage->vsn, idfp, 14, 
           SENSOR, 0, num_tbls, tbls_to_apply,tbl_oper, &baz_sdd, 0, 0);
   if (rcode != ALL_OKAY){
      printf("PAD_ERROR: failed to convert_to_units b_az\n");
      return PEACE_PAD_ERROR;
   }
} else {
   printf("PAD_ERROR: Problem with result from read_drec iel sens 14\n");
   return PEACE_PAD_ERROR;
}
*b_az = baz_sdd;

/* sensor 15 is B-field polar bin (unitless) 
*  forward file after read */	

ret_val = read_drec (dkey, lineage->exten, lineage->vsn, idfp, 15,fwd, 0);
if (ret_val < 0){
            printf ("\nError %d from read_drec. iel sens 15\n", ret_val);
            return PEACE_PAD_ERROR;
}
if (ret_val == ALL_OKAY || idfp->filled_data){
   rcode = convert_to_units (dkey, lineage->exten, lineage->vsn, 
      idfp, 15, SENSOR, 0,num_tbls, tbls_to_apply,tbl_oper, &bpol_sdd, 0, 0);
   if (rcode != ALL_OKAY){
      printf("PAD_ERROR: failed to convert_to_units b_pol\n");
      return PEACE_PAD_ERROR;
   }
} else {
   printf("PAD_ERROR: Problem with result from read_drec iel sens 15\n");
   return PEACE_PAD_ERROR;
}
*b_pol_iel = bpol_sdd;


/* Let's check to see what we've got: */
if(DIAG_PAD)printf("b_az = %g; b_pol_iel = %d\n", *b_az,*b_pol_iel);

return ALL_OKAY;
}


/************************************************************************/
SDDAS_SHORT set_b_para_half_spin(float b_az, 
          IDFS_LINEAGE *lngp, PARALLEL_FIELD_HALF_SPIN *b_para_half_spin){
/************************************************************************/
/* set the half spin which contains the +B field LOOK direction.
*
*  Note 1: b_az is value returned in PEACE CORE FGM IEL data. According
*  to various email exchanges, this is the angle through which LEEA
*  rotates from start of spin to be looking in +B direction.
*
*  Note 1a: [June 2002] Since b_az is related to the field at the END
*  of the spin, it is possible for it to be in a different half of the
*  spin from that predicted at the spin start. In this case, the 3
*  azimuths selected on board will be from the wrong spin half; since
*  one of these is returned, this causes the algorithm described below
*  to be in error. I believe the actual azimuthal slice returned is
*  also telemetred somewhere (e.g., in calibrationn data). But this may
*  be in terms of azimuthal slice NUMBER, and so is mode dependent.
*
*  Note 2: FGM is located 203.5 degrees spinward of LEEA. So if azimuth
*  of B in FGM system is az_fgm = arc tan Bz_fgm/By_fgm then LEEA must
*  rotate az_fgm + 203.5 degrees to be looking in +B. According to
*  Paul Carter (email 8 March 2001) THIS is what is put in the IEL, i.e.,
*
*  b_az = 203.5 + arctan(Bz_fgm/By_fgm)   % 360
*
*  Thus this is the azimuth relative to LEEA, and NOT relative to sun
*  direction, at start of spin.
*
*  [There was further email discussion, so this may not be the last
*  word, but seems correct to me.]
*
*  Note 3: FGM IEL Vir Inst also include Bpolbin. This is the polar zone
*     related to B. The zone is computed by calculating
*     arctan (Bx_fgm/sqrt(By_fgm**2 + Bz_fgm**2)) [actually this is
*               the LATITUDE, not polar angle]
*     Then the instrument zone is chosen which contains this latitude. 
*     If b_az > 180, the polar zone is flipped to ( 11 - this zone)
*     This gives the polar zone containe the LINE of B in the 1st half
*     spin. And this value is returned in the FGM IEL data.
*
*     This is not needed here, as I sort the pa bins myself, and determine
*     empirically (by reading the data) the 1st zone filled. But Bpolbin
*     does provide a useful check.
*
*  Algorithm is therefore simple:
*
*      For LEEA:  if b_az < 180: +B is contained in 1st half spin
*                         > 180: ......             2nd ....
*
*      For HEEA:  opposite of above:
*
*   Note 4:  PEACE actually stores on board 3 contiguous azimuths and then
*      selects one to transmit. However, it ALWAYS stores 3 azimuths
*      from the same half spin. E.g., if there are 16 az sectors 0, 1, 2, ...
*      then if sector 7 is predicted, then 5, 6, 7 are stored. This ensures
*      that the algorithm at the end of the preceding note works.
*
*   Note 5.  B_az and Bpolbin are time-tagged with the spin AFTER the
*      one in which the PAD is measured. The same is NOT true for the
*      field components Bx, By, Bz.
*
*   Note 6: There appears to be a timing error in that the time from
*      the start of the spin to the 1st half-spin of pad data is
*      returned divided by TWO. This means it is at an azimuth which is
*      likewise incorrect. The 2nd half spin is 180 degrees from the first;
*      that is, their relative positions are correct, but they don't match
*      B_az. This has not been confirmed by MSSL, though they understand
*      the problem and are working 'round it by defining an entirely new
*      VIDF. But this appears to be the case. Perhaps it will be corrected.
*      Once this has been established, however, it is straightforward to
*      check it if need be, or simply ignore it, as the information in B_az
*      and Bpolbin (and/or the 1st sensor filled with data) are sufficient
*      to reconstruct the instrument's behaviour.
*/

if (b_az < 0. || b_az > 360.){
   printf("b_az = %g out of range in set_b_para_half_spin\n", b_az);
   return PEACE_PAD_ERROR;
}

if (strcmp(lngp->virtual_inst,"CPPADL")==0){
   *b_para_half_spin = (b_az<=180) ? PARALLEL_FIRST_HALF : PARALLEL_SECOND_HALF;
} else if (strcmp(lngp->virtual_inst,"CPPADH")==0){
   *b_para_half_spin = (b_az>180) ? PARALLEL_FIRST_HALF : PARALLEL_SECOND_HALF;
} else {
     printf("set_b_para_spin_half Virt Inst = %s, not CPPADL or CPPADH\n",
                   lngp->virtual_inst);
     return PEACE_PAD_ERROR;
}

return ALL_OKAY;
}



/*********************************************************************/
int compute_clockbin(int sensor, int num_sensors, int swp_half, int n_pa){
/*********************************************************************/
/* return position 'round full polar clock of sensor
*  uses index of swp_half (0 or 1). */

int itemp;

    if (swp_half == 0){
       return sensor;
    } else if (swp_half == 1){
       itemp = 2*num_sensors -1 -sensor;
       if (itemp < 0 || itemp > 2*n_pa -1){
          printf("PAD ERROR: Clock pos = %d not >0 or < %d\n",itemp,
	          2*n_pa-1);
          return -1;
       }
       return itemp;
    } else {
       printf("Invalid swp_half = %d in compute_clockbin\n",swp_half);
       return -1;
    }

}

/******************************************************************/
SDDAS_SHORT sort_pa_bins(int *inbins, int *outbins, int pa_dim,
    PARALLEL_FIELD_HALF_SPIN bpara_half, int b_polbin){
/******************************************************************/
/*
*  inbins[i] contains the clock values for the pa_dim sweeps found
*  outbins[j] will contain the indices i in pa order, with
*     outbins[0] = i <=> field-aligned FLOW 
*     outbins[pa_dim -1] = i <=> anti-field-aligned FLOW
*
*  Algorithm:
*     create a full clock array (clockbins) and initialise to -1
*     take each inbin[i] and put its "i" in clockbin[inbin[i]]
*     find field-aligned clockbin and determine which way 'round to go
*     fill outbins[i] in order from clockbin
*/

int n_clockbins, *clockbins= NULL;
int b_aligned_clockbin;
CLOCK_DIRECTION step;  /* +1 for clockwise, -1 for anti-clockwise, 0 for error */

int i,j;

n_clockbins = 2*(pa_dim -1);   /* eg pa_dim = 13 <=> 24 clockbins */

clockbins = malloc(n_clockbins*sizeof(int));

for (i=0; i< n_clockbins; i++) clockbins[i] = -1;

for (i=0; i< pa_dim; i++){
   j = inbins[i];
   if (j < 0 || j >= n_clockbins){
       printf("sort_pa_bins inbin[%d] = %d out of range for %d clockbins\n",
                              i,j,n_clockbins);
       free(clockbins);
       return PEACE_PAD_ERROR;
   } else {
       clockbins[j] = i;
   }
}  /* at this stage, clockbin[j] contains index i s.t. inbins[i] = j */

if(DIAG_PAD){
   printf("i  inbins[i]\n");
   for (i=0; i<pa_dim; i++){
      if (i%6 == 0) printf("\n");
      printf("(%d  %d) ",i,inbins[i]);
   }
   printf("\ni clockbins[i]\n");
   for (i=0; i<n_clockbins; i++){
      if (i%6 == 0) printf("\n");
      printf("(%d  %d) ",i,clockbins[i]);
   }
   printf("\n");
}
/* 
*  flip tip and tail of B to find peace FLOW direction <=> para to B
*  find b_aligned clockbin. Note number of sensors (e.g. 12) is pa_dim -1 
*
*  NB This commented out May 2001, so that everything works in LOOK
*  directions. This is consistent with def'n of Sweep structure, and
*  also (tbc) with convention used by new SPINPAD data 
*
*  NB NB This uncommented June 2002 so that everything works in FLOW
*  directions corresponding to up-dated def'n of PEACE_SWEEP
*/
b_aligned_clockbin = exchange_b_tip_tail(&b_polbin,&bpara_half,pa_dim -1);
if (b_aligned_clockbin < 0){
    printf("exchange_b_tip_tail returned error\n");
    free(clockbins);
    return PEACE_PAD_ERROR;
}

b_aligned_clockbin = compute_clockbin(b_polbin,pa_dim-1, bpara_half,pa_dim);
if (b_aligned_clockbin < 0){
    printf("compute_clockbin returned error for b_aligned_clockbin\n");
    free(clockbins);
    return PEACE_PAD_ERROR;
}

if(DIAG_PRINT)printf("b_aligned_clockbin  = %d; sweep half = %d\n",
         b_aligned_clockbin, bpara_half);

/* determine the direction to proceed from b_aligned bin, and also try to nudge
*  b_aligned bin if it isn't on a boundary between filled (>=0) and empty (<0)
*  regions around the clock
*/
step = clockstep(clockbins, &b_aligned_clockbin, n_clockbins, (n_clockbins/2)+1);
if (step == 0){
   printf("Clockstep returned 0 to sort_pa_bins\n");
   free(clockbins);
   return PEACE_PAD_ERROR;
}
/* use clockstep results to fill outbins */

for (i=0, j=b_aligned_clockbin + n_clockbins; i<pa_dim; i++, j+=step){
    outbins[i] = clockbins[j % n_clockbins];
    if (outbins[i] < 0){
       printf("sort_pa_bins tried to put empty clockbin[%d] into outbin[%d]\n",
             i, j % n_clockbins);
       free(clockbins);
       return PEACE_PAD_ERROR;
    }
}

if(DIAG_PRINT){
   printf("i  outbins[i]\n");
   for (i=0; i<pa_dim; i++){
      if (i%6 == 0) printf("\n");
      printf("(%d  %d) ",i,outbins[i]);
   }
}

free(clockbins);
return ALL_OKAY;

}

/********************************************************************/
int exchange_b_tip_tail(int *bpol, PARALLEL_FIELD_HALF_SPIN *bpara_half, 
                             int n_sensors){
/*******************************************************************/
/* given presumed information on the PEACE polar bin and spin half
*  for the b-field, choose the opposite direction. This may need
*  doing either to get it right or to convert from look to flow
*  directions - or both
*/

if (*bpara_half == PARALLEL_FIRST_HALF)*bpara_half = PARALLEL_SECOND_HALF;
else *bpara_half = PARALLEL_FIRST_HALF;

*bpol = n_sensors -1 - *bpol;

if (*bpol < 0 || *bpol > n_sensors -1) return PEACE_PAD_ERROR;
else return 0;
}

/*****************************************************************/
CLOCK_DIRECTION clockstep(int *clockbins, int *startbin, int nbins, 
                   int tries_to_go){
/*****************************************************************/
/* return step (+1 or -1) required to step from field-aligned to anti-aligned
*  bins based on info in clockbins.
*
*  returns zero on error.
*
*  startbin should be bounded on one side by a filled bin, and on the other
*  by an empty one. If it isn't, it is nudged and we try again recursivly for
*  a while. This actual startbin is returned through the function arg
*/

if(DIAG_PRINT)printf("clockstep: startbin: %d, nbins: %d, tries_to_go: %d\n",
             *startbin, nbins, tries_to_go);

if (--tries_to_go < 0){
    printf("clockstep ran out of tries\n");
    return CLOCKSTEPERROR;
}
if (*startbin > nbins + 2) *startbin -= nbins;
if (*startbin < 2) *startbin += nbins;  /* will work modulo nbins; avoid negatives*/

if (clockbins[*startbin % nbins] >= 0){    /* startbin has data */
   if ( clockbins[(*startbin + 1) % nbins] < 0 ) return COUNTERCLOCKWISE;
   else if (clockbins[(*startbin -1) % nbins] < 0) return CLOCKWISE;
}

/* if here, then orig startbin not on boundary between filled and unfilled
*  nudge it by one and try again - oh no, not some RECURSION! */

(*startbin)++;

return clockstep(clockbins, startbin, nbins, tries_to_go);

}

/*******************************************************************/
PEACE_SWEEP* PADLiouvilleMap (PEACE_SWEEP *swp1, float B1, float B2, 
  float epot, int n_theta_double){
/*****************************************************************
*
*  Use Liouville's theorem assuming adiabatic particle behaviour to
*  map a PAD of data in a PEACE_SWEEP structure. Formal mapping moves
*  pitch angles in energy-dependent way, so formally mapped bins aren't
*  rectangular. Here we :
*
*  1  Increase energy bin energies by potential
*  2  Re-grid/interpolate the original data onto a finer theta mesh
*        Input number of times to double theta resolution; 7 is a good
*        number (128) for orig bins of ~15 degrees. Limits 1-10
*  3  Map Re-gridded bin centres using Liouville's Thm
*  4  Locate appropriate bin and ADD f(v) value there
*  5  Keep track of how many vals end up in same bin to take avg
*
*  The finer mesh allows us to use bin centres rather than try to
*  to calculate distorted areas.
*
*  Basic Liouville Map:
*
*  W2 = W1 + epot 
*  sin^2 theta2 = (W1/W2) (B2/B1) sin^2 theta1 <---use mid-pt bin W's
*
*  If sin^2 theta2 > 1 <=> mirrored particle. 
*
*  Inputs:
*     swp1   Empty sweep pointer
*     B1     Magnetic field magnitude <=> swp1 [only B2/B1 matters]
*     B2     Magnetic field magnitude <=> "mapped to" sweep
*     epot   e(Phi2 - Phi1) in eV
*
*  Output:
*     swp2   Mapped distribution (NULL if error)
*
*/


PEACE_SWEEP *swp0r, *swp1r,*swp2;
int nen1, nen1r, nen2, ien, nth1, nth1r, nth2, ith;
float *weights, *thlo1r, *thhi1r, *thlo2, *thhi2, *emid1, *emid2;
float v2efac = 2.8429e-6;    /* E(eV) = v2efac * [v(km/s)]^2 */
                             /* v2efac = m_e x 1.e6 / (2 x e) */
float missing;  /* added nov 2002 */
float epotvsq, elotmp1, elotmp2, ehitmp1, ehitmp2;
float brat, thmid1, sinsqth2map, thmap; 
int i, inbin, iz1, iz2;
char vinstr[LINEAGE_MAX_STR];

switch (swp1->psd_units){
   case PSD_UNITS_S3_PER_KM6:      /* only f(v) units at present */
      break;                       /* continue with mapping */
   default:
      printf("Liouville Map Error: only units <=> f(v) allowed\n");
      return NULL;
}

if (B1 == 0.0){
   printf("Liouville Map Error: B1 = 0!\n");
   return NULL;
} else {
   brat = fabs(B2/B1);
}

nen1 = swp1->n_lev;
nth1 = swp1->n_sen;

/********************************************* increase theta resolution */

if (n_theta_double < 1) n_theta_double = 1;  /* protection against silly input*/
if (n_theta_double > 10)n_theta_double = 10; /* protection against silly input*/

swp1r = create_sweep_full_copy(swp1);  /* work on a copy */
if (swp1r == NULL){
    printf("Liouville Map: couldn't create swp1r copy\n");
    return NULL;
}

for (i=0; i<n_theta_double; i++){
   swp0r = swp1r;                              /* point to previous sweep */
   swp1r = sweep_double_sensors(swp0r);        /* double resolution       */
   free_sweep_arrays(swp0r); 
   free(swp0r); swp0r=NULL;                    /* throw away prev. double */
}

nen1r = swp1r->n_lev;
nth1r = swp1r->n_sen;

/* now swp1r should contain unmapped data but interpolated onto a finer
*  angle mesh. Make a copy of original sweep on coarser scale 
* to hold mapped data. */

swp2 = create_sweep_full_copy(swp1);
if (swp2 == NULL){
     printf("Liouville Map Error: couldn't create swp2 copy\n");
     free_sweep_arrays(swp1r); free(swp1r);
     return NULL;
}

nen2 = swp2->n_lev;
nth2 = swp2->n_sen;
missing = swp2->missing_value;      
weights = malloc (nen2 * nth2 * sizeof(float));
emid1 = malloc(nen2*sizeof(float));
emid2 = malloc(nen2*sizeof(float));
for (i=0; i< nen2 * nth2; i++){ weights[i] = 0.0; swp2->psd[i] = missing; } 
thlo1r = swp1r->thlo;    /* local copies */
thhi1r = swp1r->thhi;
thlo2  = swp2->thlo;
thhi2  = swp2->thhi;

switch (swp2->bin_units){   /* Energy bins: shift & calc mid-pts*/
  case BIN_UNITS_EV:
    for (ien = 0; ien < nen2; ien++){
      emid1[ien] = 0.5 *(swp1r->elo[ien] + swp1r->ehi[ien]); /* mid-bin E1 */
      swp2->elo[ien] += epot;                 /* shift energy bins */
      swp2->ehi[ien] += epot;
      emid2[ien] = 0.5 *(swp2->elo[ien] + swp2->ehi[ien]); /* mid-bin E2 */
     }
     break;
   case BIN_UNITS_KM_S:  /* convert epot to (km/s)^2; */
      epotvsq = epot/v2efac;
      for (ien = 0; ien < nen2; ien++){
        elotmp1 = swp1r->elo[ien];   elotmp1 *=  elotmp1;
	elotmp2 = elotmp1 + epotvsq;  
	swp2->elo[ien] = sqrt(elotmp2);
        ehitmp1 = swp1r->ehi[ien];   ehitmp1 *=  ehitmp1;
	ehitmp2 = ehitmp1 + epotvsq;  
	swp2->ehi[ien] = sqrt(ehitmp2);
	emid1[ien] = 0.5 *(elotmp1 + ehitmp1);
	emid2[ien] = 0.5 *(elotmp2 + ehitmp2);
      }
      break;
   default:
       printf("Liouville Map Error: Unknown bin units\n");
       free_sweep_arrays(swp1r); free(swp1r);
       free_sweep_arrays(swp2); free(swp2);
       free(weights); free(emid1); free(emid2);
       return NULL;
}


/* now do pangle mapping; note this is energy-dependent; place f(v) into
*  resulting pangle bin. If sin^2 > 1.0 then skip <=> mirrored*/

for (ith = 0; ith < nth1r; ith++){               /* pangle map: */
  thmid1 = 0.5 *( thlo1r[ith] + thhi1r[ith]);
  for (ien = 0; ien < nen1r; ien ++){
     sinsqth2map = sinsq_map(emid1[ien],emid2[ien],brat,thmid1);
     if (!(sinsqth2map < 0.0 || sinsqth2map > 1.0 )){
	thmap = (asin( sqrt(sinsqth2map)))/DEG2RAD;
	if (thmid1 > 90.) thmap = 180.0 - thmap;
	if (thmap >= thlo2[0] && thmap <= thhi2[nth2-1]){
	  inbin = 0;                            /* find new angle bin */
	  while (thhi2[inbin] < thmap && inbin < nth2-1)inbin++;
	  iz1 = nen1r*ith + ien;   /* psd[ith][ien] = psd[nen*ith+ien]*/
	  iz2 = nen2*inbin+ien;
	  if (swp1r->psd[iz1] != missing){ /* skip if old missing */
	    if (swp2->psd[iz2] == missing){
	       swp2->psd[iz2] = swp1r->psd[iz1];
	    } else {
	        swp2->psd[iz2] += swp1r->psd[iz1]; /* collect in new bin */
            }
	    weights[iz2]   += 1.0;
	 }
      }
    }
  }
}

for (iz2 = 0; iz2 < nen2*nth2; iz2++){
   if (weights[iz2] > 0.0) swp2->psd[iz2] /= weights[iz2];
}

free_sweep_arrays(swp1r); free(swp1r);
free(weights); free(emid1); free(emid2);

sprintf(vinstr, "%s B%4.2g->%4.2g,e\\gf%4.2geV",
                swp2->lineage->virtual_inst,B1, B2, epot);
strncpy(swp2->lineage->virtual_inst, vinstr, LINEAGE_MAX_STR -1);
/*strcpy (swp2->lineage->experiment,"");
strcpy (swp2->lineage->instrument,""); */  /* now use sup & supsuptitle*/

printf("PADLiouvilleMap returning ...\n");
return swp2;

}

/********************************************************************/
PEACE_SWEEP *sweep_double_sensors(PEACE_SWEEP *swpin){
/*******************************************************************/
/*  Increase the number of sensors by factor 2 with new overlapping
*   bins to simulate interpolation 
*
*   Input    swpin    filled sweep structure
*
*   Returns  pointer to new sweep structure*/

int nen1, nth1, ien, ith, nen1r, nth1r;
PEACE_SWEEP bin_holder, *swpout;
SDDAS_SHORT ret_val;

nen1 = swpin->n_lev;
nth1 = swpin->n_sen;

/* increase angle res. by fac 2. Do this by using rebin_sweep routine
*  with a new angle grid WHICH IS STAGGERED wrt original and finer. E.g.
*
*               0      1       2       3         nth1 = 4
*  Old:    |- - - -|- - - -|- - - -|- - - -|
*  New:    |-|- -|- -|- -|- -|- -|- -|- -|-|
*           0  1   2   3   4   5   6   7  8      nth2 = 9
*
*  This guarantees some smoothing. */

bin_holder.n_lev =  nen1r= nen1;     /* no change in energy levels */
bin_holder.n_sen  = nth1r= 2*nth1 + 1;   /* see diag above */  
bin_holder.bin_units = swpin->bin_units;

ret_val = malloc_sweep_arrays(&bin_holder, QPFALSE); /* no way to keep pa */
if (ret_val != ALL_OKAY){
  printf("sweep_double_sensors Error: couldn't malloc refined sweep arrays\n");
  return NULL;
}
bin_holder.missing_value = swpin->missing_value;

for (ien = 0; ien< nen1; ien++){
   bin_holder.elo[ien] = swpin->elo[ien];
   bin_holder.ehi[ien] = swpin->ehi[ien];
}
bin_holder.thlo[0] = swpin->thlo[0];   /* make new th bins */
for (ith = 0; ith < nth1-1; ith++){  
   bin_holder.thlo[2*ith + 1] = 
                0.25*(3.0 * swpin->thlo[ith] +     swpin->thlo[ith+1]);   
   bin_holder.thlo[2*ith + 2] = 
                0.25*(    swpin->thlo[ith] + 3.0 * swpin->thlo[ith+1]);   
}
bin_holder.thlo[2*nth1 -1] = 0.25*(3.0*swpin->thlo[nth1-1]+swpin->thhi[nth1-1]);
bin_holder.thlo[2*nth1]    = 0.25*(swpin->thlo[nth1-1]+3.0*swpin->thhi[nth1-1]);
for (ith = 0; ith < nth1r -1 ; ith ++){
    bin_holder.thhi[ith] = bin_holder.thlo[ith+1];
}
bin_holder.thhi[nth1r-1] = swpin->thhi[nth1-1];

swpout = rebin_sweep(&bin_holder, swpin); 
free_sweep_arrays(&bin_holder);

return swpout;

}

/********************************************************************/
PEACE_SWEEP *sweep_double_levels(PEACE_SWEEP *swpin){
/*******************************************************************/
/*  Increase the number of levels by factor 2 with new overlapping
*   bins to simulate interpolation 
*
*   Input    swpin    filled sweep structure
*
*   Returns  pointer to new sweep structure*/

int nen1, nth1, ien, ith, nen1r, nth1r;
PEACE_SWEEP bin_holder, *swpout;
SDDAS_SHORT ret_val;

nen1 = swpin->n_lev;
nth1 = swpin->n_sen;

/* increase energy res. by fac 2. Do this by using rebin_sweep routine
*  with a new angle grid WHICH IS STAGGERED wrt original and finer. E.g.
*
*               0      1       2       3         nen1 = 4
*  Old:    |- - - -|- - - -|- - - -|- - - -|
*  New:    |-|- -|- -|- -|- -|- -|- -|- -|-|
*           0  1   2   3   4   5   6   7  8      nen2 = 9
*
*  This guarantees some smoothing. */

bin_holder.n_lev =  nen1r= 2*nen1+1;     /* see diag above */
bin_holder.n_sen  = nth1r= nth1;   /* no change to th bins */  
bin_holder.bin_units = swpin->bin_units;

ret_val = malloc_sweep_arrays(&bin_holder, QPFALSE); /* no way to keep pa */
if (ret_val != ALL_OKAY){
  printf("sweep_double_levels Error: couldn't malloc refined sweep arrays\n");
  return NULL;
}
bin_holder.missing_value = swpin->missing_value;

for (ith = 0; ith< nth1; ith++){
   bin_holder.thlo[ith] = swpin->thlo[ith];
   bin_holder.thhi[ith] = swpin->thhi[ith];
}
bin_holder.elo[0] = swpin->elo[0];   /* make new energy bins */
for (ien = 0; ien < nen1-1; ien++){  
   bin_holder.elo[2*ien + 1] = 
                0.25*(3.0 * swpin->elo[ien] +     swpin->elo[ien+1]);   
   bin_holder.elo[2*ien + 2] = 
                0.25*(    swpin->elo[ien] + 3.0 * swpin->elo[ien+1]);   
}
bin_holder.elo[2*nen1 -1] = 0.25*(3.0*swpin->elo[nen1-1]+swpin->ehi[nen1-1]);
bin_holder.elo[2*nen1]    = 0.25*(swpin->elo[nen1-1]+3.0*swpin->ehi[nen1-1]);
for (ien = 0; ien < nen1r -1 ; ien ++){
    bin_holder.ehi[ien] = bin_holder.elo[ien+1];
}
bin_holder.ehi[nen1r-1] = swpin->ehi[nen1-1];

swpout = rebin_sweep(&bin_holder, swpin); 
free_sweep_arrays(&bin_holder);

return swpout;

}

/********************************************************************/
PEACE_SWEEP *PADvpara_boost(PEACE_SWEEP *swpin, float vparaboost,
   int thetafactor, int levelfactor){
/********************************************************************/
/* Boost a PAD along the parallel axis by vparaboost (in km/s)
*
*  Inputs:
*     swpin        the sweep to work on
*     vparaboost   the boost velocity; all pts move from v// to v//+this
*     thetafactor  increase angular resolution by this amount
*     levelfactor  increase level resolution by this amount
*
*     thetafactor = 2, levelfactor = 4 work pretty well
*       [the routine interpolates initially to four times this, adds
*         vparaboost, then decreases to these] 
*  Algorithm:
*
*  Increase the number of pitch angles and energy levels to smooth
*      and interpolate original sweep
*  Create empty sweep and initialise weights to zero
*  Convert to velocity if in energy
*  For each old bin
*      Calculate centre as mid(v^2), mid(theta)
*      Add vparallel to find new centre
*      Locate new bin
*      Add to old contents and add one to weight
*  Normalise all values by weight
*  Convert bins to energy if orig bins in energy
*  Return
*
*  Notes:
*   1  We don't use any  area weights, etc. to avoid complications.
*      Since the vparallel part doesn't change the resolution, this
*      simplification should have little impact, as we expect only 1
*      or two entries per bin. [The initial re-binning to increase
*      resolution and extend energy range DOES use weighted re-binning.]
*
*/
PEACE_SWEEP  *swp2pa2en,*swpout, *swpout2,bin_holder;
int npa, nlev, npsd;  /* output dims */
int i, ipa, ilev, iold, inew, thbin, vbin;
float *weights, *vmid, *thmid;        /* use midpts */
float *thlo, *thhi, *elo, *ehi, *psdout, *psdin; /* local copies of things*/
float v2efac = 2.8429e-6;    /* E(eV) = v2efac * [v(km/s)]^2 */
                             /* v2efac = m_e x 1.e6 / (2 x e) */
float missing, vparanew, vperp, vnew, thnew,emin,emax;
char instr[LINEAGE_MAX_STR];

/* ************************************ re-bin and extend energy range */

nullify_sweep_arrays(&bin_holder);
bin_holder.n_sen =  4*thetafactor*swpin->n_sen;
bin_holder.n_lev =  4*levelfactor*swpin->n_lev;
bin_holder.bin_units = swpin->bin_units;
bin_holder.thlo = (float *)malloc(bin_holder.n_sen * sizeof(float));
bin_holder.thhi = (float *)malloc(bin_holder.n_sen * sizeof(float));
bin_holder.elo = (float *)malloc(bin_holder.n_lev * sizeof(float));
bin_holder.ehi = (float *)malloc(bin_holder.n_lev * sizeof(float));
if (bin_holder.thlo == NULL || bin_holder.thhi == NULL 
  || bin_holder.elo == NULL || bin_holder.ehi  == NULL){
  printf("PADvparaboost failed to malloc bin_holder\n");
  free_sweep_arrays(&bin_holder);
  return NULL;
}
set_equal_bins_half_ends(bin_holder.thlo, bin_holder.thhi,
                   bin_holder.n_sen, 0.0, 180.0);
switch(swpin->bin_units){
   case BIN_UNITS_KM_S:
      emin = swpin->elo[0]-fabs(vparaboost);
      emax = swpin->ehi[swpin->n_lev -1]+fabs(vparaboost);
      if (emin<0.0) {emin = 0.1*swpin->elo[0];}
      break;
   case BIN_UNITS_EV: default:
      emin = sqrt(swpin->elo[0]/v2efac) - fabs(vparaboost);
      if (emin < 0.0) {emin = 0.1*swpin->elo[0];}
      else {emin = v2efac*emin*emin;}
      emax = sqrt(swpin->ehi[swpin->n_lev -1]/v2efac)+fabs(vparaboost);
      emax = v2efac*emax*emax;
      break;
}
set_equal_bins_linlog(bin_holder.elo, bin_holder.ehi, 
    bin_holder.n_lev, emin, emax, PGSPECTRO_LOG);
swp2pa2en = rebin_sweep(&bin_holder, swpin); 
free_sweep_arrays(&bin_holder);
/********************************************** initialize *****************/
npa = swp2pa2en->n_sen;
nlev = swp2pa2en->n_lev;
npsd = npa*nlev;
weights = (float *)malloc(npa*nlev*sizeof(float));
vmid    = (float *)malloc(nlev*sizeof(float));
thmid   = (float *)malloc(npa* sizeof(float));
swpout = create_sweep_full_copy(swp2pa2en);  /* initialise everything */
missing = swpout->missing_value;
if (weights == NULL || vmid == NULL || thmid == NULL || swpout == NULL){
   printf("PADvpara_boost create/malloc error\n");
   if (weights != NULL) free(weights);
   if (vmid != NULL) free(vmid);
   if (thmid != NULL) free(thmid);
   if (swpout != NULL) {free_sweep_arrays(swpout); free(swpout);}
   if (swp2pa2en != NULL) {free_sweep_arrays(swp2pa2en); free(swp2pa2en);}
   return NULL;
}
elo  = swpout->elo;   ehi = swpout->ehi;     /* local ptr copies */
thlo = swpout->thlo;  thhi = swpout->thhi;
psdout  = swpout->psd;
psdin   = swp2pa2en->psd;
for (i=0; i<npsd; i++){
   weights[i] = 0.0;
   psdout[i]  = missing;
}

/****************************** MID-POINTS *****************************/
/* always use energy midpoint (since polar area prop to v^2)
*  BUT convert here to km/s to ease mapping*/
switch(swpout->bin_units){
   case BIN_UNITS_KM_S:
      for(i=0; i<nlev; i++)
        vmid[i] = sqrt(0.5*(elo[i]*elo[i]+ehi[i]*ehi[i]));
      break;
   case BIN_UNITS_EV:
      for(i=0; i<nlev; i++){
        vmid[i]=sqrt(0.5*(elo[i]+ehi[i])/v2efac);
	elo[i] = sqrt(elo[i]/v2efac);  /* convert energy to v(km/s) */
	ehi[i] = sqrt(ehi[i]/v2efac);  /* need to convert back at end */
      }
      break;
   default:
      printf("Unrecognised bin units in PADvpara_boost\n");
      printf("Defaulting to energy - use with caution\n");	  
      for(i=0; i<nlev; i++){
        vmid[i]=sqrt(0.5*(elo[i]+ehi[i])/v2efac);
	elo[i] = sqrt(elo[i]/v2efac);  /* convert energy to v(km/s) */
	ehi[i] = sqrt(ehi[i]/v2efac);  /* need to convert back at end */
      }
      break;
}
for (i=0; i<npa; i++)thmid[i] = 0.5*(thlo[i]+thhi[i]);

/************************************ ready to go to work ************/
for (ipa = 0; ipa < npa; ipa++){               
  for (ilev = 0; ilev < nlev; ilev ++){
     vparanew = vmid[ilev]*cos(DEG2RAD*thmid[ipa]) + vparaboost;
     vperp = vmid[ilev]*sin(DEG2RAD*thmid[ipa]);
     vnew = sqrt(vparanew*vparanew + vperp*vperp);
     thnew = (acos(vparanew/vnew))/DEG2RAD;
 
     thbin = inbin(thnew,thlo,thhi,npa);
     vbin  = inbin(vnew, elo, ehi, nlev);
     if (thbin >=0 && vbin >=0){
        iold = nlev*ipa + ilev;   /* psd[ith][ien] = psd[nen*ith+ien]*/
	inew = nlev*thbin + vbin;
	if (psdin[iold] != missing){
	    if (psdout[inew] == missing) {psdout[inew] = psdin[iold];}
	    else {psdout[inew] += psdin[iold];}
	    weights[inew] += 1.0;
	}
     }	
  }
}

for (inew = 0; inew < nlev*npa; inew++){
   if (weights[inew] > 0.0) swpout->psd[inew] /= weights[inew];
}

sprintf(instr, "%s V//+=%4.2gkm/s",
                swpout->lineage->instrument,vparaboost);
strncpy(swpout->lineage->instrument, instr, LINEAGE_MAX_STR -1);
/*strcpy (swpout->lineage->experiment,"");
strcpy (swpout->lineage->instrument,""); */ /* now use sup and supsuptitle */

/* if bin units should be energy, convert back from velocity */

switch(swpout->bin_units){
   case BIN_UNITS_KM_S: break;
   case BIN_UNITS_EV: default:
      for(i=0; i<nlev; i++){
         elo[i] = v2efac*elo[i]*elo[i];
	 ehi[i] = v2efac*ehi[i]*ehi[i];
      }
      break;
}
/* now swpout contains a fine resolution sweep */

nullify_sweep_arrays(&bin_holder);   /* chop number of bins, not completely */
bin_holder.n_sen =  thetafactor*swpin->n_sen;
bin_holder.n_lev =  levelfactor*swpin->n_lev; 
bin_holder.bin_units = swpin->bin_units;
bin_holder.thlo = (float *)malloc(bin_holder.n_sen * sizeof(float));
bin_holder.thhi = (float *)malloc(bin_holder.n_sen * sizeof(float));
bin_holder.elo = (float *)malloc(bin_holder.n_lev * sizeof(float));
bin_holder.ehi = (float *)malloc(bin_holder.n_lev * sizeof(float));
if (bin_holder.thlo == NULL || bin_holder.thhi == NULL 
  || bin_holder.elo == NULL || bin_holder.ehi  == NULL){
  printf("PADvparaboost failed to malloc 2nd bin_holder\n");
  free_sweep_arrays(&bin_holder);
  free(weights); free(vmid); free(thmid);
  free_sweep_arrays(swp2pa2en); free(swp2pa2en);
  return NULL;
}
set_equal_bins_half_ends(bin_holder.thlo,bin_holder.thhi,
         bin_holder.n_sen, 0.0, 180.0);
set_equal_bins_linlog(bin_holder.elo, bin_holder.ehi, 
    bin_holder.n_lev, swpout->elo[0], swpout->ehi[swpout->n_lev-1], 
    PGSPECTRO_LOG);

swpout2 = rebin_sweep(&bin_holder, swpout); 
free_sweep_arrays(&bin_holder);  
/* clean up and exit */
free_sweep_arrays(&bin_holder);  
free(weights); free(vmid); free(thmid);
/*free_sweep_arrays(swp2pa);    free(swp2pa); */
free_sweep_arrays(swp2pa2en); free(swp2pa2en);
free_sweep_arrays(swpout); free(swpout);

printf("PADvpara_boost returning\n");

return swpout2;   

}


/********************************************************************/      
float sinsq_map (float e1, float e2, float b2ob1, float thin){
/********************************************************************/
/* use 1st adiabatic invariant to find sin^2 pitch angle in region2
*  given energies in region1 and region2 (same unless a potential),
*  ratio of field magnitude, and initial pitch angle (deg)
*
*  Returns sin^2 rather than pangle so calling routine can detect
*  for itself mirrored particles and take evasive action
*/

float sinth1, sinsq;

sinth1 = sin (DEG2RAD * thin);
sinsq = (e1/e2)*b2ob1 * sinth1 * sinth1;

return sinsq;

}


/********************************************************************/
float cyclic360abs_diff(float x1, float x2){
/********************************************************************/
/* find the absolute (unsigned) difference between two angles around a 
* 360 circle */

float diff;

diff = fabs(x1 - x2);
if (diff > 180.) diff = 360. - diff;
return diff;
}

/********************************************************************/
float cyclic360pos_diff(float xlo, float xhi){
/********************************************************************/
/* find the positive directed difference between two angles around a 
* 360 circle such that adding this difference to xlo arrives at the
* position of xhi */

while (xhi < xlo)xhi += 360. ;
return xhi - xlo;
}


#ifdef __cplusplus
}
#endif



