/* Qsweep_export.c ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*
*  May 2002 separate out from Qsweep_spin
*
*  Routines for exporting PEACE sweeps using qie modules
*
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "Qidfs_utils.h"
#include "Qsweep_spin.h"
#include "Qpeace_pad.h"

/*********************************************************************/
SDDAS_SHORT nsweeps2QiCDFContents(PEACE_SWEEP **swppp, int nswps,
                                  QiCDFContents **QiCDFp){
/*********************************************************************/
/* take nswps sweeps and create and fill a QiCDFContents structure */

SDDAS_SHORT ret;
int i, dim, nth, nlev;

QiCDFVariable_novary novary_opts[9] = {EVERY_RECORD, EVERY_RECORD, EVERY_RECORD,
                EVERY_RECORD, EVERY_RECORD, 
		EVERY_RECORD, EVERY_RECORD,
		EVERY_RECORD, EVERY_RECORD};
long sizeofentrys[9] = {sizeof(double), sizeof(double), sizeof(float),
                        sizeof(float),  sizeof(float),  
			sizeof(float),  sizeof(float),  
			sizeof(float),  sizeof(float)};
long numbers[9] = {0, 1, 2, 3, 4, 5, 6, 7, 8};
char *names[9] = {"epoch",       "epoch_end",    "psd", 
                  "phi_start",    "phi_end", 
		  "theta_start",  "theta_end",    
		  "level_start",  "level_end"};
long data_types[9] = {CDF_EPOCH,   CDF_EPOCH,  CDF_FLOAT,
                      CDF_FLOAT,   CDF_FLOAT,  
		      CDF_FLOAT,   CDF_FLOAT,   
		      CDF_FLOAT,   CDF_FLOAT};
long rec_varys[9]  = {VARY, VARY, VARY, VARY, VARY, VARY, VARY, VARY, VARY};
long num_dimss[9]   = {0, 0, 2, 0, 0, 1, 1, 1, 1};
long num_elemss[9]  = {1, 1, 1, 1, 1, 1, 1, 1, 1};

long dim_size_epoch[1] = {0},       dim_size_epoch_end[1] = {0}; 
long dim_size_psd[2];
long dim_size_phi_start[1] = {0},   dim_size_phi_end[1] = {0};
long dim_size_theta_start[1], dim_size_theta_end[1];
long dim_size_lev_start[1],   dim_size_lev_end[1];
long *dim_sizess[9] = {dim_size_epoch, dim_size_epoch_end, dim_size_psd,
       dim_size_phi_start,   dim_size_phi_end, 
       dim_size_theta_start, dim_size_theta_end, 
       dim_size_lev_start,   dim_size_lev_end};
       
long dim_varys_epoch[1]       = {VARY}, dim_varys_epoch_end[1] = {VARY};
long dim_varys_psd[2]         = {VARY, VARY}; 
long dim_varys_phi_start[1]   = {VARY}, dim_varys_phi_end[1]   = {VARY};
long dim_varys_theta_start[1] = {VARY}, dim_varys_theta_end[1] = {VARY};
long dim_varys_lev_start[1]   = {VARY}, dim_varys_lev_end[1]   = {VARY};
long *dim_varyss[9] = {dim_varys_epoch, dim_varys_epoch_end, dim_varys_psd,
      dim_varys_phi_start,   dim_varys_phi_end,   
      dim_varys_theta_start, dim_varys_theta_end,
      dim_varys_lev_start,   dim_varys_lev_end};
      
QiCDFContents *QiCDF;

nth  = swppp[0]->n_sen;
nlev = swppp[0]->n_lev;
for (i=1; i<nswps; i++){   /* check sweeps all same size */
   if (swppp[i]->n_sen != nth || swppp[i]->n_lev != nlev){
      printf("nswps2QiCDF: Sweep %d has different length from 1st sweep.\n",i);
      return PEACE_SWEEPS_SIZE_DIFFER;
   }
}
/* set up as much as possible; details, incl attrs done in separate routines*/
dim_size_psd[0]         = nth;         dim_size_psd[1]       = nlev;
dim_size_theta_start[0] = nth ;        dim_size_theta_end[0] = nth;
dim_size_lev_start[0]   = nlev;        dim_size_lev_end[0]   = nlev;

QiCDF = QiMakeCDFContentsObj();   /* ptrs all initialised to NULL or STR_NULL*/
*QiCDFp = QiCDF;                  /* pass ptr back to calling routine */
QiCDF->n_vars = 9;   /* epoch start, end,  psd, phi start, end,  */
                     /* theta start, end, levels start, end   */
QiCDF->n_recs = nswps;
QiCDF->time_var_num = 0;
QiCDF->vardata = QiMakeQiVariablePtrs(QiCDF->n_vars);

ret= nswps2QiCDF_g_atts(QiCDF, swppp);

for (i=0; i<QiCDF->n_vars; i++){ 
    QiCDF->vardata[i] = QiMakeQiVariable();
    QiCDF->vardata[i]->novary_opt = novary_opts[i];
    QiCDF->vardata[i]->sizeofentry= sizeofentrys[i];
    QiCDF->vardata[i]->number     = numbers[i];
    QiCDF->vardata[i]->name       = QiNewStr(names[i]);
    QiCDF->vardata[i]->data_type  = data_types[i];
    QiCDF->vardata[i]->rec_vary   = rec_varys[i];
    QiCDF->vardata[i]->max_rec_num= nswps-1;
    QiCDF->vardata[i]->num_dims   = num_dimss[i];
    QiCDF->vardata[i]->num_elems  = num_elemss[i];
    for (dim = 0; dim < num_dimss[i]; dim++){  /* copy into malloced QiCDF */
	QiCDF->vardata[i]->dim_sizes[dim]  = *(dim_sizess[i]+dim);
	QiCDF->vardata[i]->dim_varys[dim]  = *(dim_varyss[i]+dim);
    }
}

ret = nswps2QiCDF_time_vars(QiCDF->vardata[0], QiCDF->vardata[1],swppp, nswps);
if (ret != ALL_OKAY) return ret;
ret = nswps2QiCDF_psd_var(QiCDF->vardata[2],swppp, nswps);
if (ret != ALL_OKAY) return ret;
ret = nswps2QiCDF_phi_vars(QiCDF->vardata[3], QiCDF->vardata[4], swppp, nswps);
if (ret != ALL_OKAY) return ret;
ret = nswps2QiCDF_theta_vars(QiCDF->vardata[5],QiCDF->vardata[6],swppp, nswps);
if (ret != ALL_OKAY) return ret;
ret = nswps2QiCDF_lev_vars(QiCDF->vardata[7], QiCDF->vardata[8], swppp, nswps);
if (ret != ALL_OKAY) return ret;

for (i=0; i<QiCDF->n_vars; i++){
   if(QiVarSafe(QiCDF, i)==0){
      printf("QiCDF Variable %i failed safe test\n", i);
      return PEACE_SWEEPS_EXPORT_ERROR;
   }
}

return ret;

}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_g_atts(QiCDFContents *QiCDF, PEACE_SWEEP **swppp){
/*********************************************************************/

char lineage_str[500],gen_time[33];
QiGlobalAttribute **g_att;
time_t now;
struct tm *date;

QiCDF->num_g_attrs = 3;   /* no mandatory globals; put lineage as global */
QiCDF->g_attr = QiMakeQiGAttrPtrs(QiCDF->num_g_attrs);

g_att = QiCDF->g_attr;  /* local copy */

g_att[0] = QiMakeQiGAttr();
g_att[0]->number = 0;
g_att[0]->name   = QiNewStr("IDFS Lineage");
g_att[0]->num_entries = 1;
g_att[0]->entry = QiMakeEntries(1);
g_att[0]->entry->exists = 1;    /*  ?????????????????????*/
g_att[0]->entry->data_type = CDF_CHAR;
sprintf(lineage_str,"%s %s %s %s",
       swppp[0]->lineage->mission,
       swppp[0]->lineage->experiment,
       swppp[0]->lineage->instrument,
       swppp[0]->lineage->virtual_inst);
g_att[0]->entry->data = QiNewStr(lineage_str);
g_att[0]->entry->num_elems = strlen(g_att[0]->entry->data) ;

g_att[1] = QiMakeQiGAttr();
g_att[1]->number = 1;
g_att[1]->name   = QiNewStr("TEXT");
g_att[1]->num_entries = 3;
g_att[1]->entry = QiMakeEntries(3);
g_att[1]->entry[0].exists = 1;   /*  yes, seems to need this*/
g_att[1]->entry[0].data_type = CDF_CHAR;
g_att[1]->entry[0].data = 
   QiNewStr("PEACE Sweep Data generated by QPEACE Software");
g_att[1]->entry[0].num_elems = strlen(g_att[1]->entry[0].data) ;
g_att[1]->entry[1].exists = 1;
g_att[1]->entry[1].data_type = CDF_CHAR;
g_att[1]->entry[1].data = 
   QiNewStr("S.J.Schwartz@qmul.ac.uk; http://www.space-plasma.qmul.ac.uk");
g_att[1]->entry[1].num_elems = strlen(g_att[1]->entry[1].data);
g_att[1]->entry[2].exists = 1;   /*  yes, seems to need this*/
g_att[1]->entry[2].data_type = CDF_CHAR;
g_att[1]->entry[2].data = QiNewStr(qpeace_version_str());
g_att[1]->entry[2].num_elems = strlen(g_att[1]->entry[2].data) ;

g_att[2] = QiMakeQiGAttr();
g_att[2]->number = 2;
g_att[2]->name   = QiNewStr("Generation_date");
g_att[2]->num_entries = 1;
g_att[2]->entry = QiMakeEntries(1);
g_att[2]->entry[0].exists = 1;   /*  yes, seems to need this*/
g_att[2]->entry[0].data_type = CDF_CHAR;
now = time(NULL); date = localtime (&now);
strftime(gen_time,33,"%Y-%m-%dT%H:%M:%S.000Z Doy %j",date);
g_att[2]->entry[0].data = QiNewStr(gen_time);
g_att[2]->entry[0].num_elems = strlen(g_att[1]->entry[0].data) ;

return ALL_OKAY;
}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_default_attr(QiVarAttribute *attp, int iatt){
/*********************************************************************/
attp->number = iatt;
attp->data_type = CDF_CHAR;
/* attp->num_elems = 1; */  /* this needs to be reset to strlen of attr */
return ALL_OKAY;
}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_time_vars(QiCDFVariable *epo, QiCDFVariable *epo_end,
                                  PEACE_SWEEP **swppp, int nswps){
/**********************************************************************/
int i;
double *epodp, *epoedp;
long num_atts = 2;

/* time attributes are Value_type (done by qie), UNITS, SI_conversion */
/* number, data_type set by nswps2QiCDF_default_attr */
/* NB qie routines set Time_format attribute 
*  NB QiMakeQiVariable allocates MAX_N_ATTRS attribute structures */

char *att_names[2] = {"UNITS", "SI_conversion"};
char *att_data[2]  = {"msecs since 0AD", "1.e-3>s"};

epo->num_v_attrs= num_atts;
epo_end->num_v_attrs = num_atts;
for (i=0; i<num_atts; i++) {
  nswps2QiCDF_default_attr( &(epo->attribute[i]),i );
  epo->attribute[i].name = QiNewStr(att_names[i]);
  epo->attribute[i].data = QiNewStr(att_data[i]);
  epo->attribute[i].num_elems = strlen(epo->attribute[i].data);

  nswps2QiCDF_default_attr( &(epo_end->attribute[i]),i );
  epo_end->attribute[i].name = QiNewStr(att_names[i]);
  epo_end->attribute[i].data = QiNewStr(att_data[i]);
  epo_end->attribute[i].num_elems = strlen(epo_end->attribute[i].data);
}

epo->data     = (double *)malloc(nswps * sizeof(double));
epo_end->data = (double *) malloc(nswps * sizeof(double));

epodp  = (double *)epo->data;
epoedp = (double *)epo_end->data;

for (i=0; i<nswps; i++){
    epodp[i]     = 1000.0 * idfs_time_to_epoch_secs(&(swppp[i]->btime));
    epoedp[i]    = 1000.0 * idfs_time_to_epoch_secs(&(swppp[i]->etime));
}

return ALL_OKAY;
}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_psd_var(QiCDFVariable *psdv, 
                                  PEACE_SWEEP **swppp, int nswps){
/**********************************************************************/
int i,j,k;
int n_per_swp;
float *psdvp;
long num_atts = 12;  /* fillval added may2002; Depend_end_0 added Feb2003 */
                     /* SC_id and FIELDNAM added oct 2003  */
char fillvalstr[50];

/* array attributes for cef are Value_type,  FILLVAL, UNITS, SI_conversion,
*        Depend_1, Depend_end_1, Depend_2, Depend_end_2, 
*        Bin_description, Bin_description_A        
*        Sizes and Value_type added by qie routines */
/* number, data_type,  set by nswps2QiCDF_default_attr */

char *att_names[12] = {"FILLVAL","UNITS", "SI_conversion","Depend_end_0",
       "DEPEND_1", "DEPEND_END_1", "DEPEND_2", "DEPEND_END_2", 
       "Bin_description", "Bin_description_a","SC_id","FIELDNAM"};
char *att_data[12]  = {"dummy fillval", "dummy units", "dummy SI_conv",
   "epoch_end","theta_start", "theta_end", "level_start", "level_end",
   "Bins are polar theta vs speed/energy in an azimuthal slice",
   "Theta directions are instrument FLOW directions","dummy SC_id",
   "Phase Space Density"};

psdv->num_v_attrs= num_atts;
for (i=0; i<num_atts; i++) {
  nswps2QiCDF_default_attr( &(psdv->attribute[i]),i );
  psdv->attribute[i].name = QiNewStr(att_names[i]);
  psdv->attribute[i].data = QiNewStr(att_data[i]);
  psdv->attribute[i].num_elems = strlen(psdv->attribute[i].data);
}

QiChangeVattr("UNITS",PSD_UNITS_STRING(swppp[0]->psd_units),psdv);
QiChangeVattr("SI_conversion",
         PSD_UNITS_SI_conv_STRING(swppp[0]->psd_units),psdv);
sprintf(fillvalstr,"%g",swppp[0]->missing_value);
QiChangeVattr("FILLVAL",fillvalstr,psdv);
QiChangeVattr("SC_id", swppp[0]->lineage->mission,psdv);

n_per_swp = swppp[0]->n_sen * swppp[0]->n_lev;
psdv->data     = (float *)malloc(nswps * n_per_swp * sizeof(float));

psdvp = (float *)psdv->data;

/* fill data array. NB psd is a 2-D array, but in right order so copy as 1-D*/
for (i=0; i<nswps; i++){
    k= i*n_per_swp;
    for (j=0; j<n_per_swp; j++){
        psdvp[k+j]  = swppp[i]->psd[j];
    }
}

return ALL_OKAY;
}


/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_phi_vars(QiCDFVariable *phiv, QiCDFVariable *phiv_end,
                                  PEACE_SWEEP **swppp, int nswps){
/**********************************************************************/
int iswp, i;
float *phivp, *phiv_endp;
long num_atts = 5;

/* Scalar variable attributes for cef are Value_type,  UNITS, SI_conversion,
*        Frame and Component_desc included for future use
*        FIELDNAM added for plots
*        Sizes and Value_type added by qie routines */
/* number, data_type  set by nswps2QiCDF_default_attr */


char *att_names[5] = {"UNITS", "SI_conversion",
       "Frame", "Component_desc", "FIELDNAM"};
char *att_data[5]  = {"Deg", "1>degree", 
    "component>pea_rtp", 
    "p>azimuthal FLOW angle from PEACE tzero","dummy attr"};

phiv->num_v_attrs= num_atts;
phiv_end->num_v_attrs= num_atts;
for (i=0; i<num_atts; i++) {
  nswps2QiCDF_default_attr( &(phiv->attribute[i]),i );
  phiv->attribute[i].name = QiNewStr(att_names[i]);
  phiv->attribute[i].data = QiNewStr(att_data[i]);
  phiv->attribute[i].num_elems = strlen(phiv->attribute[i].data);
  nswps2QiCDF_default_attr( &(phiv_end->attribute[i]),i );
  phiv_end->attribute[i].name = QiNewStr(att_names[i]);
  phiv_end->attribute[i].data = QiNewStr(att_data[i]);
  phiv_end->attribute[i].num_elems = strlen(phiv_end->attribute[i].data);
}

QiChangeVattr("FIELDNAM","Azimuth (PEACE tzero) sweep start", phiv);
QiChangeVattr("FIELDNAM","Azimuth (PEACE tzero) sweep end",  phiv_end);

phiv->data     = (float *)malloc(nswps  * sizeof(float));
phivp = (float *)phiv->data;
phiv_end->data     = (float *)malloc(nswps  * sizeof(float));
phiv_endp = (float *)phiv_end->data;

for (iswp=0; iswp<nswps; iswp++){
    phivp[iswp]   = swppp[iswp]->phi_lo;
    phiv_endp[iswp] = swppp[iswp]->phi_hi;
}

return ALL_OKAY;
}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_theta_vars(QiCDFVariable *thv, QiCDFVariable *thv_end, 
                                  PEACE_SWEEP **swppp, int nswps){
/**********************************************************************/
int iswp, jth, k, nth,  i;
float *thvp, *thv_endp;
long num_atts = 7;

/* Dimension var attributes for cef are Value_type,  UNITS, SI_conversion,
*        Bin_location, Scaling; COULD ADD Frame & Component_desc 
*        Sizes and Value_type added by qie routines */
/* number, data_type, and num_elems set by nswps2QiCDF_default_attr */


char *att_names[7] = {"UNITS", "SI_conversion",
       "Bin_location", "Scaling", "Frame", "Component_desc","FIELDNAM"};
char *att_data[7]  = {"Deg", "1>degree",
   "to be overwritten", "linear", "component>pea_rtp", "t>polar FLOW angle",
   "dummy attr"};

thv->num_v_attrs= num_atts;
thv_end->num_v_attrs= num_atts;
for (i=0; i<num_atts; i++) {
  nswps2QiCDF_default_attr( &(thv->attribute[i]),i );
  thv->attribute[i].name = QiNewStr(att_names[i]);
  thv->attribute[i].data = QiNewStr(att_data[i]);
  thv->attribute[i].num_elems = strlen(thv->attribute[i].data);
  nswps2QiCDF_default_attr( &(thv_end->attribute[i]),i );
  thv_end->attribute[i].name = QiNewStr(att_names[i]);
  thv_end->attribute[i].data = QiNewStr(att_data[i]);
  thv_end->attribute[i].num_elems = strlen(thv_end->attribute[i].data);
}

QiChangeVattr("Bin_location", "0.0", thv);
QiChangeVattr("Bin_location", "1.0", thv_end);
QiChangeVattr("FIELDNAM", "Start polar angle", thv);
QiChangeVattr("FIELDNAM", "End polar angle", thv_end);

nth = swppp[0]->n_sen;

thv->data     = (float *)malloc(nswps * nth * sizeof(float));
thvp = (float *)thv->data;
thv_end->data     = (float *)malloc(nswps * nth * sizeof(float));
thv_endp = (float *)thv_end->data;

/* data is data[nswp][nth] so data[i][j] = data[i*nth + j] */

for (iswp=0; iswp<nswps; iswp++){
    k = iswp*nth;       /* indx of 1st elem in data for swpp[iswp] */
    for (jth = 0; jth < nth; jth++){
       thvp[k+jth]   = swppp[iswp]->thlo[jth];
       thv_endp[k+jth] = swppp[iswp]->thhi[jth];
    }
}

return ALL_OKAY;
}

/*********************************************************************/
SDDAS_SHORT nswps2QiCDF_lev_vars(QiCDFVariable *levv, QiCDFVariable *levv_end,
                                  PEACE_SWEEP **swppp, int nswps){
/**********************************************************************/
int iswp, jlev, k,  nlev, i;
float *levvp, *levv_endp;
long num_atts = 7;

/* array attributes for cef are Value_type,  UNITS, SI_conversion,
*        Bin_location, Scaling; COULD ADD Frame & Component_desc 
*        Sizes and Value_type added by qie routines */
/* number, data_type, and num_elems set by nswps2QiCDF_default_attr */


char *att_names[7] = {"UNITS", "SI_conversion",
       "Bin_location", "Scaling", "Frame", "Component_desc","FIELDNAM"};
char *att_data[7]  = {"dummy units", "dummy SI_conversion",
      "to be over-written", "linear", "component>pea_rtp", "dummy comp",
      "dummy name"};

levv->num_v_attrs= num_atts;
levv_end->num_v_attrs= num_atts;
for (i=0; i<num_atts; i++) {
  nswps2QiCDF_default_attr( &(levv->attribute[i]),i );
  levv->attribute[i].name = QiNewStr(att_names[i]);
  levv->attribute[i].data = QiNewStr(att_data[i]);
  levv->attribute[i].num_elems = strlen(levv->attribute[i].data);
  nswps2QiCDF_default_attr( &(levv_end->attribute[i]),i );
  levv_end->attribute[i].name = QiNewStr(att_names[i]);
  levv_end->attribute[i].data = QiNewStr(att_data[i]);
  levv_end->attribute[i].num_elems = strlen(levv_end->attribute[i].data);
}

QiChangeVattr("Bin_location","0.0",levv);
QiChangeVattr("Bin_location","1.0",levv_end);
QiChangeVattr("FIELDNAM","Start Level",levv);
QiChangeVattr("FIELDNAM","End Level",levv_end);

QiChangeVattr("UNITS",BIN_UNITS_STRING(swppp[0]->bin_units),levv);
QiChangeVattr("SI_conversion",
         BIN_UNITS_SI_conv_STRING(swppp[0]->bin_units),levv);
QiChangeVattr("UNITS",BIN_UNITS_STRING(swppp[0]->bin_units),levv_end);
QiChangeVattr("SI_conversion",
         BIN_UNITS_SI_conv_STRING(swppp[0]->bin_units),levv_end);

switch(swppp[0]->bin_units){
    case BIN_UNITS_EV :
      QiChangeVattr("Component_desc","E>energy",levv); 
      QiChangeVattr("Component_desc","E>energy",levv_end); 
      break;
    case BIN_UNITS_KM_S :
      QiChangeVattr("Component_desc","v>speed",levv); 
      QiChangeVattr("Component_desc","v>speed",levv_end);
      break;
    default :
      QiChangeVattr("Component_desc","Level>unknown measure",levv); 
      QiChangeVattr("Component_desc","Level>unknown measure",levv_end);
      break;
}

nlev = swppp[0]->n_lev;

levv->data     = (float *)malloc(nswps * nlev * sizeof(float));
levvp = (float *)levv->data;
levv_end->data     = (float *)malloc(nswps * nlev * sizeof(float));
levv_endp = (float *)levv_end->data;

/* data is data[nswp][nlev] so data[i][j = data[i*nlev + j] */

for (iswp=0; iswp<nswps; iswp++){
    k = iswp*nlev;       /* indx of 1st elem in data for swpp[iswp] */
    for (jlev = 0; jlev < nlev; jlev++){
       levvp[k+jlev]     = swppp[iswp]->elo[jlev];
       levv_endp[k+jlev] = swppp[iswp]->ehi[jlev];
    }
}

return ALL_OKAY;
}

#ifdef __cplusplus
}
#endif
