/* qjas.cc ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  last up-date 16 July 2001
*  March 2002 - implemented zero as missing value
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opeining and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/


#include <stdio.h>

#include "qjas.h"

 /*
 * Constructor
 *
 * Creates child widgets of the main QjasBox widget
 */

QjasBox::QjasBox( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    setMargin( 5 );
    setSpacing( 5 );

    QString str;

    QHBox *row1 = new QHBox( this );
    row1->setSpacing( 5 );

    // Create a LineageBox which has comboboxes for the various lineage components
    lineage = new LineageBox( row1 );
 
    QVBox *timecol = new QVBox (row1);
       timecol->setSpacing(5);
       QLabel *timecol_label = new QLabel("Start and End Times", timecol);
       timecol_label->setAlignment(AlignLeft);
    starttime = new TimeBox(1, Horizontal, timecol);

    QHBox *copy_time_box  = new QHBox(timecol);
    QLabel *copy_time_label = new QLabel("Copy Time   ", copy_time_box);
       copy_time_label->setAlignment( AlignRight | AlignVCenter);
    QPushButton *copy_time = new QPushButton ("|\nV",copy_time_box);
    new QLabel("     ",copy_time_box);
    connect (copy_time, SIGNAL (clicked() ),
             this , SLOT (slotCopyTime() ) );

    endtime = new TimeBox(1, Horizontal, timecol);

    QHBox *scale_opts = new QHBox (this);
    scale_opts->setSpacing(5);
    QVBox *scale_labels       = new QVBox (scale_opts);
    	                        scale_labels->setSpacing(5);
    QVBox *scale_units        = new QVBox (scale_opts);
    	                        scale_units->setSpacing(5);
    QVBox *scale_linlog       = new QVBox (scale_opts);
    	                        scale_linlog->setSpacing(5);
    QVBox *scale_auto         = new QVBox (scale_opts);
    	                        scale_auto->setSpacing(5);
    QVBox *scale_minmax_label = new QVBox (scale_opts);
    	                        scale_minmax_label->setSpacing(5);
    QVBox *scale_min          = new QVBox (scale_opts);
    	                        scale_min->setSpacing(5);
    QVBox *scale_max          = new QVBox (scale_opts);
     	                        scale_max->setSpacing(5);

// Energy/velocity options
    new QLabel ("Energy/Velocity Options", scale_labels);
    energy_units = new QComboBox(FALSE,scale_units);
       for (int ibu = 0; ibu < N_BIN_UNITS; ibu++){
          energy_units->insertItem (BIN_UNITS_STRING((BIN_UNITS)ibu), -1);
       }
    energy_bins = new QComboBox(FALSE,scale_linlog);
       energy_bins->insertItem("Instrument Bins",-1);
       energy_bins->insertItem("Custom Bins",-1);
   new QLabel ("",scale_auto);
   new QLabel ("",scale_minmax_label);
   new QLabel ("",scale_min);
   new QLabel ("",scale_max);
// Custom bins options
    new QLabel ("   Custom bins: Number", scale_labels);
    nebins = new QLineEdit("10",scale_units);
    energy_linlog = new QComboBox(FALSE,scale_linlog);
       energy_linlog->insertItem("Linear",-1);
       energy_linlog->insertItem("Log",-1);
       energy_linlog->setCurrentItem(1);
   energy_scale = new QComboBox(FALSE,scale_auto);
       energy_scale->insertItem("Instrument Range",-1);
       energy_scale->insertItem("User Range ->",-1);
     new QLabel ("min, max", scale_minmax_label);
    emin = new QLineEdit ("10.00", scale_min);
    emax = new QLineEdit ("1000.00", scale_max);

// PSD Options
    new QLabel ("Phase Space Options", scale_labels);
    psd_units = new QComboBox(FALSE,scale_units);
       for (int ipsu = 0; ipsu < N_PSD_UNITS; ipsu++){
          psd_units->insertItem (PSD_UNITS_STRING((PSD_UNITS)ipsu), -1);
       }
    psd_linlog = new QComboBox(FALSE,scale_linlog);
       psd_linlog->insertItem("Linear",-1);    /* same order as PGSpectro.h */
       psd_linlog->insertItem("Log",-1);
       psd_linlog->setCurrentItem(1);
    psd_scale = new QComboBox(FALSE,scale_auto);
       psd_scale->insertItem("Autoscale",-1); /* same order as PGSpectro.h */
       psd_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    psdmin = new QLineEdit ("1.00000", scale_min);
//    psdmin->setFixedWidth(70);
    psdmin->setMinimumWidth(70);
    psdmax = new QLineEdit ("10.0000", scale_max);
//    psdmax->setFixedWidth(70);
    psdmax->setMinimumWidth(70);
    
// Plot Options
    QHBox *plot_opts = new QHBox (this);
    plot_opts->setSpacing(5);
    new QLabel ("Plot Options", plot_opts);
    plot_col = new QComboBox(FALSE,plot_opts);
       plot_col->insertItem("Rainbow",-1);   /* same order as PGSpectro.h */
       plot_col->insertItem("Gray White Low",-1);
       plot_col->insertItem("Gray White High", -1);
    plot_dev = new QComboBox(FALSE,plot_opts);
       plot_dev->insertItem("Xwindow",-1);
       plot_dev->insertItem("Landscape GIF File ->",-1);
       plot_dev->insertItem("Portrait GIF File ->",-1);
       plot_dev->insertItem("Landscape PS File ->",-1);
       plot_dev->insertItem("Portrait PS File ->",-1);
    new QLabel ("Xwin Number", plot_opts);
    plot_win_edit = new QLineEdit ("3", plot_opts);
    plot_win_edit->setMaxLength(3); plot_win_edit->setFixedWidth(45);
    new QLabel ("Plot File Stem", plot_opts);
    plot_file_edit = new QLineEdit ("qjas_plot", plot_opts);

// Export Options
    QHBox *exp_opts_box = new QHBox (this);
    exp_opts_box->setSpacing(5);
    new QLabel ("Export to File Options", exp_opts_box);
    exp_opt = new QComboBox(FALSE,exp_opts_box);
       exp_opt->insertItem("Ascii Cluster Exchange",-1);
       exp_opt->insertItem("Ascii Tabular",-1);
       exp_opt->insertItem("CDF", -1);
    new QLabel ("Export File Stem", exp_opts_box);
    exp_file_edit = new QLineEdit ("qjas_data", exp_opts_box);


// Action Buttons
    QHBox *actions = new QHBox (this);
    actions->setSpacing(5);
    QPushButton *plotbut = new QPushButton( "Plot",actions );
    QPushButton *exprtbut = new QPushButton( "Export", actions);
    QPushButton *help = new QPushButton( "Help", actions);
    connect(exprtbut,  SIGNAL(clicked()),  this, SLOT ( slotExp()) );
    connect( plotbut,  SIGNAL(clicked()) , this, SLOT (slotPlot()) );
    connect (help,     SIGNAL(clicked()) , this, SLOT (slotHelp()) );
    quit = new QPushButton( "Quit",actions);

// create help window 
    helpbwin = new Helpbwin(); 

}



/*
*     Constructor for LineageBox 
*/

LineageBox::LineageBox( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    this->setSpacing(5);
    new QLabel("IDFS Lineage", this);
    proj = new  QComboBox (FALSE, this);
    proj->insertStrList (project_strs);
    miss = new  QComboBox (FALSE, this);
    for (int isc = 0; isc < N_SC; isc++){
          miss->insertItem (SC_NAME((SC)isc), -1);
    }
    exp  = new  QComboBox (FALSE, this);
    exp->insertStrList (exp_strs);
    inst = new  QComboBox (FALSE, this);
    inst->insertStrList (inst_strs);
    virins = new  QComboBox (FALSE, this);
    virins->insertStrList (virins_strs);
    pachoice = new QComboBox (FALSE, this);
       pachoice->insertItem("no pa binning", -1);
       pachoice->insertItem("pa: angles", -1);
       pachoice->insertItem("pa: cosines", -1);    
    npa = new QLineEdit("13", this);
      QToolTip::add(npa, "Number of pitch angle bins for rebinning");
//    npa->setFixedWidth(30);

}

void LineageBox::grab_lineage(IDFS_LINEAGE *lngp){

strcpy(lngp->project,proj->currentText().latin1());
strcpy(lngp->mission,miss->currentText().latin1());
strcpy(lngp->experiment,exp->currentText().latin1());
strcpy(lngp->instrument,inst->currentText().latin1());
strcpy(lngp->virtual_inst,virins->currentText().latin1());
strcpy(lngp->exten,"");
if (DIAG_PRINT_3D)print_idfs_lineage(lngp);

}

/*
*     Constructors for TimeBox
*/

TimeBox::TimeBox (int strips, Orientation orientation,
                          QWidget *parent, const char *name )
     : QGroupBox (strips, orientation, parent, name)
{

QHBox *row_time = new QHBox (this);
row_time->setSpacing(30);
new QLabel("     ",row_time);  // space
time_lbl   = new QLabel ("yyyy doy hh:mm:ss.sss", row_time);
year = new QLineEdit ("2001", row_time);
year->setMaxLength(4);year->setFixedWidth(60);
doy = new QLineEdit ("090", row_time);
doy->setMaxLength(3);doy->setFixedWidth(45);
hr = new QLineEdit ("18", row_time);
hr->setMaxLength(2);hr->setFixedWidth(30);
min = new QLineEdit ("8", row_time);
min->setMaxLength(2);min->setFixedWidth(30);
sec = new QLineEdit ("12.345", row_time);
sec->setMaxLength(6);sec->setFixedWidth(75);
new QLabel("     ",row_time);  // space

}

void TimeBox::GetTimeFromTimeBox(IDFS_TIME *idftp){

int year, doy, hr, min, isec;
long nano;
float secs;

year = this->year->text().toInt();
doy  = this->doy->text().toInt();
hr   = this->hr->text().toInt();
min  = this->min->text().toInt();
secs = this->sec->text().toFloat();

isec = (int)secs;
nano = (long)(1000000000*(secs-isec));

to_idfs_time(year, doy, hr, min, isec, nano, idftp);
if (DIAG_PRINT_3D)print_idfs_time(idftp);

}

/* Constructor and destructor for help window */

Helpbwin::Helpbwin(QWidget *parent, const char *name) 
   : QVBox(parent,name){

resize(600,600);
setCaption("qjas help");

QHBox *buttons = new QHBox(this);
QPushButton *backbutton, *forwardbutton, *topbutton;

backbutton = new QPushButton( "<- Back", buttons, 0 );
forwardbutton = new QPushButton ("Forward ->",buttons, 0);
topbutton = new QPushButton ("Top",buttons,0);

QTextBrowser *helpbrowser = new QTextBrowser(this);

connect( backbutton, SIGNAL( clicked() ), helpbrowser, SLOT( backward() ) );
connect( forwardbutton, SIGNAL( clicked() ), helpbrowser, SLOT( forward() ) );
connect( topbutton, SIGNAL( clicked() ), helpbrowser, SLOT( home() ) );

QString homepath = QString(getenv("QPEACE_HOME")) + "/doc/";

helpbrowser->mimeSourceFactory()->setFilePath(homepath);
helpbrowser->setSource("qjas.html");

}

Helpbwin::~Helpbwin(void){

hide();
}


void QjasBox::slotCopyTime(void){
endtime->year->setText( starttime->year->text() );
endtime->doy->setText( starttime->doy->text() );
endtime->hr->setText( starttime->hr->text() );
endtime->min->setText( starttime->min->text() );
endtime->sec->setText( starttime->sec->text() );
}

void QjasBox::slotPlot(void){

SDDAS_SHORT ret_val;
int is_pad, nrebinned, nsc, nlev;
PEACE_SPIN3D **spinpp;
PEACE_SWEEP **swppp_unbin;
PEACE_SWEEP bin_holder;
float emin_val, emax_val;
QString s; 

lineage->grab_lineage(&lng1);

starttime->GetTimeFromTimeBox(&idft1);
endtime->GetTimeFromTimeBox(&idft2);

bin_units = (BIN_UNITS)energy_units->currentItem();
psdu = (PSD_UNITS)psd_units->currentItem();

PGSpectro_set_defaults(&pgspectro);

pgspectro.color_table = (PGSpectro_col)plot_col->currentItem();
pgspectro.zlinlog     = (PGSpectro_linlog)psd_linlog->currentItem();
Yes_or_No zautoscale  = (Yes_or_No)psd_scale->currentItem();
pgspectro.zscalemin        = psdmin->text().toFloat();
pgspectro.zscalemax        = psdmax->text().toFloat();
pgspectro.skip_missing     = YES;
pgspectro.zmissing         = QP_DEFAULT_MISSING; /* missing value reset below*/

QString pgfile, pgfileexten;
QString pgdev, pgdevstr;

pgfile=plot_file_edit->text(); // get file name; if xwin, reset below

switch(plot_dev->currentItem()){
   case 1:
      pgdev.setLatin1("/GIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 2:
      pgdev.setLatin1("/VGIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 3:
      pgdev.setLatin1("/CPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 4:
      pgdev.setLatin1("/VCPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 0:  default:
      bool ok;
      pgfile = plot_win_edit->text();
      int ixwin = pgfile.toInt( &ok, 10);  // see if text in pgfile ok xwin
      if (!ok || ixwin < 1 ||ixwin > 255){ // if not open default xwin
          pgfile.setLatin1(PLOTAZ_DEFAULT_XWIN_STR, -1); 
      }
      pgdev.setLatin1("/Xserv", -1);
      pgfileexten.setLatin1("");
      break;
}
char pg_dev_cstr[200];
pgdevstr = pgfile.append(pgfileexten);
pgdevstr.append(pgdev);
strcpy(pg_dev_cstr, pgdevstr.latin1());
// swoptions.pg_device =  pg_dev_cstr;

if (cpgopen(pg_dev_cstr) <=0){
 printf("Failed to open pgplot device %s successfully\n", pg_dev_cstr);
 printf("If file check write permissions; if Xwindow, check pgxwin server\n");
 return;       // early exit, before getting sweeps and malloc memory
}
cpgshls(0,0,1,0);
cpgshls(1,0,0,0);

is_pad = strcmp("PAD",lng1.instrument);
if (is_pad==0) is_pad=1; else is_pad = 0;
int pach = lineage->pachoice->currentItem();
if (pach == NO_BINNING){
   paopts.do_pa = QPFALSE;
} else {
   paopts.do_pa = QPTRUE;
   if (pach == PA_ANGLES) paopts.mu_bins = QPFALSE;
       else paopts.mu_bins = QPTRUE;
   paopts.n_pabins = lineage->npa->text().toInt();
}
swppp = swppp_unbin = NULL;
spinpp = NULL;   // for safety

if (strncmp(lng1.mission,"C1234",5)){nsc=1;}else{nsc=4;}/* single s/c plot? */

cpgsubp(1,nsc);
int isc;
for (isc = 0; isc < nsc; isc++){
    cpgpage();
    PGSpectro_domain_set_defaults(&(pgspectro.domain));
    pgspectro.wedge.disp = 2.5;  /* leave room for Flow 0 180 label */
    if (nsc == 4) {
       pgspectro.char_size = 1.4;
       strcpy(lng1.mission, SC_NAME((SC_E)isc));
    }
    if (paopts.do_pa == QPFALSE) {  
       ret_val = GetNPeaceSweeps(&lng1,&idft1,&idft2,bin_units,psdu,
               paopts.do_pa, &nswp, &swppp);
    } else {  // pa binning requested
      if (is_pad) {
	  ret_val = GetNPeaceSweeps(&lng1,&idft1,&idft2,bin_units,psdu,
        	  paopts.do_pa, &nswp, &swppp_unbin);
	  if (ret_val == ALL_OKAY){
              ret_val =  Rebin_n_sweeps_to_pa(swppp_unbin,  nswp,
                                	1, &paopts,&swppp, &nrebinned);
              free_n_sweeps(nswp, swppp_unbin); 
	      free(swppp_unbin);  swppp_unbin = NULL;// don't need these now
	      nswp = nrebinned;                   // in case some failed?
	  }
      } else {  // not PAD, so gather spins
          ret_val = GetNPeaceSpins(&lng1, &idft1, &idft2, bin_units, psdu,
	          paopts.do_pa, &nswp, &spinpp);
          if (ret_val == ALL_OKAY){
	      ret_val = Rebin_n_spins_to_pa(spinpp, nswp, &paopts, &swppp);
	      free_n_spins( spinpp, nswp); free(spinpp); 
	      spinpp = NULL;               // don't need these any more
          } else {
	     // nothing needed here as GetNPeaceSpins frees up memory on error
	     // ret_val passes through to next if
	  }
      }
    }
    if(ret_val == ALL_OKAY){
       cong_swps=NULL;
       if (energy_bins->currentItem() == 1){ // custom bins requested
         bin_holder = *(swppp[0]);           // copy contents; no mallocing
	 nlev = bin_holder.n_lev = nebins->text().toInt(); // set up ebins
         if ( energy_scale->currentItem() == 1){            // user scale
	     emin_val = emin->text().toFloat();
	     emax_val = emax->text().toFloat();
	 } else {
	     emin_val = swppp[0]->elo[0];
	     emax_val = swppp[0]->ehi[swppp[0]->n_lev -1];
	 }
         bin_holder.elo = (float *)malloc(nlev*sizeof(float));
	 bin_holder.ehi = (float *)malloc(nlev*sizeof(float));
	 set_equal_bins_linlog(bin_holder.elo, bin_holder.ehi, nlev, 
	            emin_val, emax_val, 
		    (PGSpectro_linlog)energy_linlog->currentItem());
	 ret_val = rebin_n_sweeps(&bin_holder, nswp, swppp, 
	                              &nrebinned, &cong_swps);
	 if (ret_val == ALL_OKAY){
	    if(zautoscale == YES){
	      sweeps_autoscaler(cong_swps, nswp, 
                	    &(pgspectro.zscalemin), &(pgspectro.zscalemax)); 
	    }
            pgspectro.zmissing = cong_swps[0]->missing_value; 
	    Qjas_plot(cong_swps, nrebinned, &pgspectro, idft1, idft2);
	    emin->setText(s.setNum(cong_swps[0]->elo[0]));/* set ebins in gui */
	    emax->setText(s.setNum(cong_swps[0]->ehi[ cong_swps[0]->n_lev -1]));
	 } else {
	    printf("Qjas tried and failed to rebin in energy\n");
	 }
	 free(bin_holder.elo); free(bin_holder.ehi);
	 free_n_sweeps(nrebinned, cong_swps); free(cong_swps); cong_swps=NULL;
       } else {  // no energy rebinning requested
	 if(zautoscale == YES){
	   sweeps_autoscaler(swppp, nswp, 
                	 &(pgspectro.zscalemin), &(pgspectro.zscalemax)); 
	 }
         pgspectro.zmissing = swppp[0]->missing_value; 
	 pg_ret_val=Qjas_plot(swppp, nswp, &pgspectro, idft1, idft2);
	 emin->setText(s.setNum(swppp[0]->elo[0]));  /* set ebins in gui */
	 emax->setText(s.setNum(swppp[0]->ehi[ swppp[0]->n_lev -1 ]));
	 if (pg_ret_val == PGSPECTRO_QJAS_INCONGRUENT){
	   cong_swps = (PEACE_SWEEP **)malloc(nswp * sizeof (PEACE_SWEEP *));
	   for (int i=0; i<nswp; i++){
	      cong_swps[i] = rebin_sweep(swppp[0], swppp[i]);
	   }
           pgspectro.zmissing = cong_swps[0]->missing_value; 
	   pg_ret_val=Qjas_plot(cong_swps, nswp, &pgspectro, idft1, idft2);
	   emin->setText(s.setNum(cong_swps[0]->elo[0])); /* set ebins in gui */
	   emax->setText(s.setNum(cong_swps[0]->ehi[cong_swps[0]->n_lev -1]));
	   free_n_sweeps(nswp,cong_swps);  free(cong_swps);    cong_swps=NULL;
	 }
       }
       psdmin->setText(s.setNum(pgspectro.zscalemin));
       psdmax->setText(s.setNum(pgspectro.zscalemax));
       free_n_sweeps(nswp,swppp);       free(swppp);        swppp=NULL;
     } else {
	printf("GetNPeaceSweeps returned error %d. Try another source/time?\n",
	    ret_val);
     }
}
cpgclos();   /* make sure pgplot dev is properly closed */
}  /* end slotPlot */

/*********************************************** also add code to rebin energies
*****************************************************************************/

void QjasBox::slotExp(void){

SDDAS_SHORT ret_val;
QiCDFContents *QiCDF;
QiOptions *QiOpts;
long check;
QString expfile;
int nswp, nlev;
PEACE_SWEEP **swppp, **cong_swps, bin_holder;

int is_pad, nrebinned;
PEACE_SPIN3D **spinpp;
PEACE_SWEEP **swppp_unbin;
float emin_val, emax_val;


QiCDF = NULL;
QiOpts = NULL;

lineage->grab_lineage(&lng1);

starttime->GetTimeFromTimeBox(&idft1);
endtime->GetTimeFromTimeBox(&idft2);

bin_units = (BIN_UNITS)energy_units->currentItem();
psdu = (PSD_UNITS)psd_units->currentItem();

is_pad = strcmp("PAD",lng1.instrument);
if (is_pad==0) is_pad=1; else is_pad = 0;
int pach = lineage->pachoice->currentItem();
if (pach == NO_BINNING){
   paopts.do_pa = QPFALSE;
} else {
   paopts.do_pa = QPTRUE;
   if (pach == PA_ANGLES) paopts.mu_bins = QPFALSE;
       else paopts.mu_bins = QPTRUE;
   paopts.n_pabins = lineage->npa->text().toInt();
}
swppp = swppp_unbin = NULL;
spinpp = NULL;   // for safety

if (strncmp(lng1.mission,"C1234",5)==0){ /* multi-s/c not exportable */
  printf("Multi-spacecraft export not supported\n");
  ret_val = PEACE_SWEEPS_EXPORT_ERROR;
} else {  /* looks ok; keep going */
//
// ************************************** get sweeps & bin in pa if requested
//
  if (paopts.do_pa == QPFALSE) {  
     ret_val = GetNPeaceSweeps(&lng1,&idft1,&idft2,bin_units,psdu,
             paopts.do_pa, &nswp, &swppp);
 } else {
     if (is_pad) {
	ret_val = GetNPeaceSweeps(&lng1,&idft1,&idft2,bin_units,psdu,
        	paopts.do_pa, &nswp, &swppp_unbin);
	if (ret_val == ALL_OKAY){
            ret_val =  Rebin_n_sweeps_to_pa(swppp_unbin,  nswp,
                1, &paopts,&swppp, &nrebinned);
            free_n_sweeps(nswp, swppp_unbin); 
	    free(swppp_unbin);  swppp_unbin = NULL; // don't need these any more
	    nswp = nrebinned;                   // in case some failed?
	}
     } else {  // not PAD, so gather spins
        ret_val = GetNPeaceSpins(&lng1, &idft1, &idft2, bin_units, psdu,
	        paopts.do_pa, &nswp, &spinpp);
        if (ret_val == ALL_OKAY){
	    ret_val = Rebin_n_spins_to_pa(spinpp, nswp, &paopts, &swppp);
        }
	free_n_spins( spinpp, nswp); free(spinpp); 
	spinpp = NULL;                          // don't need these any more
     }
  }
}
//
// *************************************** fill CDFContents Obj
//
if(ret_val == ALL_OKAY){ 
   cong_swps=NULL;
   if (energy_bins->currentItem() == 1){ // custom bins requested
     bin_holder = *(swppp[0]);           // copy contents; no mallocing
     nlev = bin_holder.n_lev = nebins->text().toInt(); // set up ebins
     if ( energy_scale->currentItem() == 1){            // user scale
	 emin_val = emin->text().toFloat();
	 emax_val = emax->text().toFloat();
     } else {
	 emin_val = swppp[0]->elo[0];
	 emax_val = swppp[0]->ehi[swppp[0]->n_lev -1];
     }
     bin_holder.elo = (float *)malloc(nlev*sizeof(float));
     bin_holder.ehi = (float *)malloc(nlev*sizeof(float));
     set_equal_bins_linlog(bin_holder.elo, bin_holder.ehi, nlev, 
	        emin_val, emax_val, 
		(PGSpectro_linlog)energy_linlog->currentItem());
     ret_val = rebin_n_sweeps(&bin_holder, nswp, swppp, 
	                          &nrebinned, &cong_swps);
     if (ret_val == ALL_OKAY){
       ret_val = nsweeps2QiCDFContents( cong_swps, nrebinned, &QiCDF);
                                // QiCDFContentsDump(QiCDF);
       if (ret_val != ALL_OKAY){
          printf("nsweeps2QiCDFContents failed for rebinned sweeps\n");
       }
       free_n_sweeps(nrebinned, cong_swps); free(cong_swps); cong_swps=NULL;
     } // ret_val drops through if there was an error re-energy-binning
   } else {  // custom bins NOT requested.
     ret_val = nsweeps2QiCDFContents( swppp, nswp, &QiCDF);
   }
//
//************************************************  set opts and export       	  
//
   if (ret_val == ALL_OKAY){
       QiOpts = QiMakeOptionsObj();
       QiOpts->header = ATTACHED;     QiOpts->rec_numbering = NUM_OFF;
       QiOpts->priority = REPLACE;    QiOpts->object_type = TS;
       QiOpts->attr_delim = ',';      QiOpts->data_delim = ',';
       QiOpts->time_sep_attr = 'T';   QiOpts->time_sep_data = 'T';
       QiOpts->fp_display = stdout;

       // set up output file path/name - only exten left till later
       expfile=(exp_file_edit->text()).simplifyWhiteSpace(); 
                    // get file name - stem only, strip all white space; 
       char *temp_str = STR_NULL;
       temp_str = (char *)malloc ((strlen(expfile.latin1()) + 1)*sizeof(char));
       strcpy(temp_str, expfile.latin1());
 
       int ilastslash = expfile.findRev("/",-1);
       if (ilastslash < 0){   // no slash found
           QiCDF->io_f_path = STR_NULL;
	   QiCDF->io_f_name = QiNewStr(temp_str);
       } else {
           QiCDF->io_f_name = QiNewStr(temp_str+ilastslash +1);
	   temp_str[ilastslash+1] = '\0';
	   QiCDF->io_f_path = QiNewStr(temp_str);
       }
       if (temp_str != STR_NULL) free(temp_str);
       
       printf("io_f_path = |%s|, io_f_name = |%s|\n",
           QiCDF->io_f_path, QiCDF->io_f_name);

       switch(exp_opt->currentItem()){
	  case 2:      //cdf
	     check = QiWriteCSDSgenCDF(QiCDF,QiOpts);
	     if (check != QMW_OK){
        	     printf("QiWriteCSDSgenCDF returned %ld\n", check);
		     printf("%s",QiErrStr(check));
	     }
	     break;
	    case 1:      // qft
	     QiOpts->f_type    = TABULAR;
	     check = QiWriteCSDSgenFlat(QiCDF,QiOpts);
	     if (check != QMW_OK){
        	     printf("QiWriteCSDSgenFlat returned %ld\n", check);
		     printf("%s",QiErrStr(check));
	     }
	     break;
	  case 0: default:     // cef
	     QiOpts->f_type    = EXCHANGE; // formerly DELIMITED;
	     QiOpts->attr_delim = ',';
             QiOpts->data_delim = ',';
             QiOpts->rec_end = '$';
             QiOpts->row_end = '\n';
	     check = QiWriteCSDSgenFlat(QiCDF,QiOpts);
	     if (check != QMW_OK){
        	     printf("QiWriteCSDSgenFlat returned %ld\n", check);
		     printf("%s",QiErrStr(check));
	     }
	     break;
       }  // end switch on exp_opt->currentItem

   } else {   // problem putting sweeps into CDFContents Obj
     printf("nsweeps2QiCDFContents returned %d\n", ret_val);
   }

   QiFreeOptionsObj (QiOpts);    // should be safe
   QiFreeCDFContentsObj( QiCDF); // ditto 
   free_n_sweeps(nswp,swppp);     
   free(swppp);      
   swppp=NULL;

} else {
   printf("GetNPeaceSweeps returned error %d. Try another source/time?\n",
       ret_val);
}

}


void QjasBox::slotHelp(void){
if ( QApplication::desktop()->width() < 600
            || QApplication::desktop()->height() < 600 ){
   helpbwin->showMaximized();
} else {
   helpbwin->show();  
}
}


